from __future__ import annotations

import decimal
import logging
import os
import re
import uuid

import ispyb.sqlalchemy as models
import marshmallow.fields
import sqlalchemy
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema
from sqlalchemy.orm import selectinload, sessionmaker

logger = logging.getLogger("cryoemservices.services.dispatcher_tools")

Session = sessionmaker(
    bind=sqlalchemy.create_engine(models.url(), connect_args={"use_pure": True})
)

re_visit_base = re.compile(r"^(.*/([a-z][a-z][0-9]+-[0-9]+))/")


def setup_marshmallow_schema(session):
    # https://marshmallow-sqlalchemy.readthedocs.io/en/latest/recipes.html#automatically-generating-schemas-for-sqlalchemy-models
    for class_ in models.Base.registry._class_registry.values():
        if hasattr(class_, "__tablename__"):

            class Meta:
                model = class_
                sqla_session = session
                load_instance = True
                include_fk = True

            TYPE_MAPPING = SQLAlchemyAutoSchema.TYPE_MAPPING.copy()
            TYPE_MAPPING.update({decimal.Decimal: marshmallow.fields.Float})
            schema_class_name = "%sSchema" % class_.__name__
            schema_class = type(
                schema_class_name,
                (SQLAlchemyAutoSchema,),
                {
                    "Meta": Meta,
                    "TYPE_MAPPING": TYPE_MAPPING,
                },
            )
            setattr(class_, "__marshmallow__", schema_class)


def get_reprocessing_info(message, parameters, session: sqlalchemy.orm.session.Session):
    reprocessing_id = parameters.get(
        "ispyb_reprocessing_id", parameters.get("ispyb_process")
    )
    if reprocessing_id:
        parameters["ispyb_process"] = reprocessing_id
        query = (
            session.query(models.ProcessingJob)
            .options(
                selectinload(models.ProcessingJob.ProcessingJobParameters),
                selectinload(models.ProcessingJob.ProcessingJobImageSweeps)
                .selectinload(models.ProcessingJobImageSweep.DataCollection)
                .load_only(
                    models.DataCollection.imageDirectory,
                    models.DataCollection.fileTemplate,
                ),
            )
            .filter(models.ProcessingJob.processingJobId == reprocessing_id)
        )
        rp = query.first()
        if not rp:
            logger.warning(f"Reprocessing ID {reprocessing_id} not found")
        # ispyb_reprocessing_parameters is the deprecated method of
        # accessing the processing parameters
        parameters["ispyb_reprocessing_parameters"] = {
            p.parameterKey: p.parameterValue for p in rp.ProcessingJobParameters
        }
        # ispyb_processing_parameters is the preferred method of
        # accessing the processing parameters
        processing_parameters: dict[str, list[str]] = {}
        for p in rp.ProcessingJobParameters:
            processing_parameters.setdefault(p.parameterKey, [])
            processing_parameters[p.parameterKey].append(p.parameterValue)
        parameters["ispyb_processing_parameters"] = processing_parameters
        schema = models.ProcessingJob.__marshmallow__()
        parameters["ispyb_processing_job"] = schema.dump(rp)
        if "ispyb_dcid" not in parameters:
            parameters["ispyb_dcid"] = rp.dataCollectionId

    return message, parameters


def get_dc_info(dcid: int, session: sqlalchemy.orm.session.Session):

    query = session.query(models.DataCollection).filter(
        models.DataCollection.dataCollectionId == dcid
    )
    dc = query.first()
    if dc is None:
        return {}
    schema = models.DataCollection.__marshmallow__()
    return schema.dump(dc)


def get_beamline_from_dcid(dcid: int, session: sqlalchemy.orm.session.Session):
    query = (
        session.query(models.BLSession)
        .join(models.DataCollectionGroup)
        .join(models.DataCollection)
        .filter(models.DataCollection.dataCollectionId == dcid)
    )
    bl_session = query.first()
    if bl_session:
        return bl_session.beamLineName


def get_visit_directory_from_image_directory(directory):
    """/dls/${beamline}/data/${year}/${visit}/...
    -> /dls/${beamline}/data/${year}/${visit}"""
    if not directory:
        return None
    visit_base = re_visit_base.search(directory)
    if not visit_base:
        return None
    return visit_base.group(1)


def get_visit_from_image_directory(directory):
    """/dls/${beamline}/data/${year}/${visit}/...
    -> ${visit}"""
    if not directory:
        return None
    visit_base = re_visit_base.search(directory)
    if not visit_base:
        return None
    return visit_base.group(2)


def dc_info_to_working_directory(dc_info):
    directory = dc_info.get("imageDirectory")
    if not directory:
        return None
    visit = get_visit_directory_from_image_directory(directory)
    rest = directory[len(visit) + 1 :]

    collection_path = dc_info["imagePrefix"] or ""
    dc_number = dc_info["dataCollectionNumber"] or ""
    if collection_path or dc_number:
        collection_path = f"{collection_path}_{dc_number}"
    return os.path.join(visit, "tmp", "zocalo", rest, collection_path, dc_info["uuid"])


def dc_info_to_results_directory(dc_info):
    directory = dc_info.get("imageDirectory")
    if not directory:
        return None
    visit = get_visit_directory_from_image_directory(directory)
    rest = directory[len(visit) + 1 :]

    collection_path = dc_info["imagePrefix"] or ""
    dc_number = dc_info["dataCollectionNumber"] or ""
    if collection_path or dc_number:
        collection_path = f"{collection_path}_{dc_number}"
    return os.path.join(visit, "processed", rest, collection_path, dc_info["uuid"])


def ready_for_processing(
    message, parameters, session: sqlalchemy.orm.session.Session | None = None
):
    """Check whether this message is ready for templatization."""

    if session is None:
        session = Session()

    if not parameters.get("ispyb_wait_for_runstatus"):
        return True

    dcid = parameters.get("ispyb_dcid")
    if not dcid:
        return True

    query = session.query(models.DataCollection.runStatus).filter(
        models.DataCollection.dataCollectionId == dcid
    )
    return query.scalar() is not None


def ispyb_filter(
    message,
    parameters,
    session: sqlalchemy.orm.session.Session | None = None,
):
    """Do something to work out what to do with this data..."""

    if session is None:
        session = Session()

    setup_marshmallow_schema(session)

    message, parameters = get_reprocessing_info(message, parameters, session)

    if "ispyb_dcid" not in parameters:
        return message, parameters

    dc_id = parameters["ispyb_dcid"]

    dc_info = get_dc_info(dc_id, session)
    if not dc_info:
        raise ValueError(f"No database entry found for dcid={dc_id}: {dc_id}")
    dc_info["uuid"] = parameters.get("guid") or str(uuid.uuid4())
    parameters["ispyb_beamline"] = get_beamline_from_dcid(dc_id, session)

    parameters["ispyb_preferred_datacentre"] = "cs05r"
    parameters["ispyb_preferred_scheduler"] = "slurm"

    parameters["ispyb_dc_info"] = dc_info

    parameters["ispyb_visit"] = get_visit_from_image_directory(
        dc_info.get("imageDirectory")
    )
    parameters["ispyb_visit_directory"] = get_visit_directory_from_image_directory(
        dc_info.get("imageDirectory")
    )
    parameters["ispyb_working_directory"] = dc_info_to_working_directory(dc_info)
    parameters["ispyb_results_directory"] = dc_info_to_results_directory(dc_info)

    if (
        "ispyb_processing_job" in parameters
        and parameters["ispyb_processing_job"]["recipe"]
        and not message.get("recipes")
        and not message.get("custom_recipe")
    ):
        # Prefix recipe name coming from ispyb/synchweb with 'ispyb-'
        message["recipes"] = ["ispyb-" + parameters["ispyb_processing_job"]["recipe"]]

    return message, parameters



from __future__ import annotations

import datetime
import subprocess
import tarfile
import time
from pathlib import Path

import htcondor
from workflows.services.common_service import CommonService

from cryoemservices.services.tomo_align import TomoAlign


class TomoAlignIris(TomoAlign, CommonService):
    """
    A service for grouping and aligning tomography tilt-series with Newstack and AreTomo
    """

    # Logger name
    _logger_name = "cryoemservices.services.tomo_align_iris"

    def parse_tomo_output(self, tomo_output_file):
        tomo_file = open(tomo_output_file, "r")
        lines = tomo_file.readlines()
        for line in lines:
            if line.startswith("Rot center Z"):
                self.rot_centre_z_list.append(line.split()[5])
            if line.startswith("Tilt offset"):
                self.tilt_offset = float(line.split()[2].strip(","))
            if line.startswith("Best tilt axis"):
                self.alignment_quality = float(line.split()[5])
        tomo_file.close()

    def aretomo(self, tomo_parameters):
        """
        Run AreTomo on output of Newstack
        """
        args = [
            "./AreTomo_1.3.0_Cuda112_09292022",
            "-OutMrc",
            tomo_parameters.aretomo_output_file,
            "-InMrc",
            str(Path(tomo_parameters.stack_file).name),
        ]

        if tomo_parameters.angle_file:
            args.extend(("-AngFile", tomo_parameters.angle_file))
        else:
            args.extend(
                (
                    "-TiltRange",
                    str(tomo_parameters.input_file_list[0][1]),  # lowest tilt
                    str(tomo_parameters.input_file_list[-1][1]),
                )
            )  # highest tilt

        if tomo_parameters.manual_tilt_offset:
            args.extend(
                (
                    "-TiltCor",
                    str(tomo_parameters.tilt_cor),
                    str(tomo_parameters.manual_tilt_offset),
                )
            )
        elif tomo_parameters.tilt_cor:
            args.extend(("-TiltCor", str(tomo_parameters.tilt_cor)))

        aretomo_flags = {
            "vol_z": "-VolZ",
            "out_bin": "-OutBin",
            "tilt_axis": "-TiltAxis",
            "flip_int": "-FlipInt",
            "flip_vol": "-FlipVol",
            "wbp": "-Wbp",
            "align": "-Align",
            "roi_file": "-RoiFile",
            "patch": "-Patch",
            "kv": "-Kv",
            "align_file": "-AlnFile",
            "align_z": "-AlignZ",
            "pix_size": "-PixSize",
            "init_val": "-initVal",
            "refine_flag": "-refineFlag",
            "out_imod": "-OutImod",
            "out_imod_xf": "-OutXf",
            "dark_tol": "-DarkTol",
        }

        for k, v in tomo_parameters.dict().items():
            if v and (k in aretomo_flags):
                args.extend((aretomo_flags[k], str(v)))

        self.log.info(f"Running AreTomo with args: {args}")
        self.log.info(
            f"Input stack: {tomo_parameters.stack_file} \n"
            f"Output file: {tomo_parameters.aretomo_output_file}"
        )

        # Set-up condor config
        htcondor.param["RELEASE_DIR"] = "/usr"
        htcondor.param["LOCAL_DIR"] = "/var"
        htcondor.param["RUN"] = "/var/run/condor"
        htcondor.param["LOG"] = "/var/log/condor"
        htcondor.param["LOCK"] = "/var/lock/condor"
        htcondor.param["SPOOL"] = "/var/lib/condor/spool"
        htcondor.param["EXECUTE"] = "/var/lib/condor/execute"
        htcondor.param["BIN"] = "/usr/bin"
        htcondor.param["LIB"] = "/usr/lib64/condor"
        htcondor.param["INCLUDE"] = "/usr/include/condor"
        htcondor.param["SBIN"] = "/usr/sbin"
        htcondor.param["LIBEXEC"] = "/usr/libexec/condor"
        htcondor.param["SHARE"] = "/usr/share/condor"
        htcondor.param["PROCD_ADDRESS"] = "/var/run/condor/procd_pipe"
        htcondor.param[
            "JAVA_CLASSPATH_DEFAULT"
        ] = "/usr/share/condor /usr/share/condor/scimark2lib.jar ."
        htcondor.param["CONDOR_HOST"] = "pool-gpu-htcondor-manager.diamond.ac.uk"
        htcondor.param["COLLECTOR_HOST"] = "pool-gpu-htcondor-manager.diamond.ac.uk"
        htcondor.param["ALLOW_READ"] = "*"
        htcondor.param["ALLOW_WRITE"] = "*"
        htcondor.param["ALLOW_NEGOTIATOR"] = "*"
        htcondor.param["ALLOW_DAEMON"] = "*"
        htcondor.param["SEC_DEFAULT_AUTHENTICATION_METHODS"] = "FS_REMOTE, PASSWORD"
        htcondor.param[
            "SEC_WRITE_AUTHENTICATION_METHODS"
        ] = "FS_REMOTE, PASSWORD, ANONYMOUS"
        htcondor.param[
            "SEC_READ_AUTHENTICATION_METHODS"
        ] = "FS_REMOTE, PASSWORD, ANONYMOUS"
        htcondor.param["FS_REMOTE_DIR"] = "/dls/tmp/htcondor"

        output_file = self.alignment_output_dir + "/" + self.stack_name + "_iris_out"
        error_file = self.alignment_output_dir + "/" + self.stack_name + "_iris_error"
        log_file = self.alignment_output_dir + "/" + self.stack_name + "_iris_log"

        try:
            at_job = htcondor.Submit(
                {
                    "executable": "/dls/ebic/data/staff-scratch/murfey/aretomo.sh",
                    "arguments": "\"'$(input_args)'\"",  # needs to be in single quotes to be interpreted as one command
                    "output": "$(output_file)",
                    "error": "$(error_file)",
                    "log": "$(log_file)",
                    "request_gpus": "1",
                    "request_memory": "10240",
                    "request_disk": "10240",
                    "should_transfer_files": "yes",
                    "transfer_input_files": "$(stack_file), /dls/ebic/data/staff-scratch/murfey/AreTomo_1.3.0_Cuda112_09292022",
                    "initialdir": "$(initial_dir)",
                }
            )
        except Exception:
            self.log.warn("Couldn't connect submitter")
            return subprocess.CompletedProcess(
                args="", returncode=1, stderr=b"IRIS couldn't connect"
            )

        if tomo_parameters.out_imod:
            itemdata = [
                {
                    "input_args": " ".join(args),
                    "stack_file": tomo_parameters.stack_file,
                    "output_file": output_file,
                    "error_file": error_file,
                    "log_file": log_file,
                    "initial_dir": self.alignment_output_dir,
                }
            ]
        else:
            itemdata = [
                {
                    "input_args": " ".join(args),
                    "stack_file": tomo_parameters.stack_file,
                    "output_file": output_file,
                    "error_file": error_file,
                    "log_file": log_file,
                    "initial_dir": self.alignment_output_dir,
                }
            ]
        coll = htcondor.Collector(htcondor.param["COLLECTOR_HOST"])
        schedd_ad = coll.locate(htcondor.DaemonTypes.Schedd)
        schedd = htcondor.Schedd(schedd_ad)
        job = schedd.submit(at_job, itemdata=iter(itemdata))
        cluster_id = job.cluster()
        self.log.info(f"Submitting to Iris, ID: {str(cluster_id)}")

        res = 1
        start_time = datetime.datetime.now()
        while res:
            try:
                res = schedd.query(
                    constraint="ClusterId=={}".format(cluster_id),
                    projection=["JobStatus"],
                )[0]["JobStatus"]
            except IndexError:
                break
            if res == 12:
                schedd.act(htcondor.JobAction.Remove, f"ClusterId == {cluster_id}")
                return subprocess.CompletedProcess(
                    args="", returncode=res, stderr=b"IRIS job failed"
                )
            time.sleep(10)
            if (datetime.datetime.now() - start_time).seconds > 20 * 60:
                # Abort if duration over 20 minutes
                schedd.act(htcondor.JobAction.Remove, f"ClusterId == {cluster_id}")
                return subprocess.CompletedProcess(
                    args="", returncode=5, stderr=b"IRIS timeout"
                )

        if tomo_parameters.tilt_cor:
            self.parse_tomo_output(output_file)

        tar_imod_dir = str(Path(self.imod_directory).with_suffix(".tar.gz"))
        file = tarfile.open(tar_imod_dir)
        file.extractall(self.alignment_output_dir)
        file.close()

        return subprocess.CompletedProcess(args="", returncode=0, stderr=b"")



from __future__ import annotations

import datetime
import json
import logging
import math
import os
from pathlib import Path
from typing import Optional

import requests
import workflows.recipe
import zocalo.configuration
from importlib_metadata import entry_points
from pydantic import BaseModel, Field
from workflows.services.common_service import CommonService
from zocalo.util import slurm


class JobSubmissionParameters(BaseModel):
    scheduler: str = "slurm"
    cluster: Optional[str]
    partition: Optional[str]
    prefer: Optional[str]
    job_name: Optional[str]
    environment: Optional[dict[str, str]] = None
    cpus_per_task: Optional[int] = None
    tasks: Optional[int] = None
    nodes: Optional[int]
    memory_per_node: Optional[int] = None
    gpus_per_node: Optional[str] = None
    min_memory_per_cpu: Optional[int] = Field(
        None, description="Minimum real memory per cpu (MB)"
    )
    time_limit: Optional[datetime.timedelta] = None
    gpus: Optional[int] = None
    exclusive: bool = False
    account: Optional[str]
    commands: str | list[str]
    qos: Optional[str]


def submit_to_slurm(
    params: JobSubmissionParameters,
    working_directory: Path,
    logger: logging.Logger,
    zc: zocalo.configuration,
) -> int | None:
    api = slurm.SlurmRestApi.from_zocalo_configuration(zc)

    script = params.commands
    if not isinstance(script, str):
        script = "\n".join(script)
    script = f"#!/bin/bash\n. /etc/profile.d/modules.sh\n{script}"

    logger.debug(f"Submitting script to Slurm:\n{script}")
    if params.time_limit:
        time_limit_minutes = math.ceil(params.time_limit.total_seconds() / 60)
    else:
        time_limit_minutes = None
    job_submission = slurm.models.JobSubmission(
        script=script,
        job=slurm.models.JobProperties(
            partition=params.partition,
            prefer=params.prefer,
            name=params.job_name,
            cpus_per_task=params.cpus_per_task,
            tasks=params.tasks,
            nodes=[params.nodes, params.nodes] if params.nodes else params.nodes,
            gpus_per_node=params.gpus_per_node,
            memory_per_node=params.memory_per_node,
            environment=os.environ
            if params.environment is None
            else params.environment,
            memory_per_cpu=params.min_memory_per_cpu,
            time_limit=time_limit_minutes,
            gpus=params.gpus,
            # exclusive=params.exclusive,
            account=params.account,
            current_working_directory=os.fspath(working_directory),
            qos=params.qos,
        ),
    )
    try:
        response = api.submit_job(job_submission)
    except requests.HTTPError as e:
        logger.error(f"Failed Slurm job submission: {e}\n" f"{e.response.text}")
        return None
    if response.errors:
        error_message = "\n".join(f"{e.errno}: {e.error}" for e in response.errors)
        logger.error(f"Failed Slurm job submission: {error_message}")
        return None
    return response.job_id


class ClusterSubmission(CommonService):
    """A service to start new jobs on a slurm cluster."""

    # Human readable service name
    _service_name = "EMCluster"

    # Logger name
    _logger_name = "cryoemservices.services.cluster"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schedulers = {}

    def initializing(self):
        """Subscribe to the cluster submission queue.
        Received messages must be acknowledged."""
        self.log.info("Cluster submission service starting")

        self.schedulers = {
            f.name: f.load()
            for f in entry_points(group="cryoemservices.services.cluster.schedulers")
        }
        self.log.debug(f"Supported schedulers: {', '.join(self.schedulers.keys())}")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "cluster.submission",
            self.run_submit_job,
            acknowledgement=True,
            log_extender=self.extend_log,
        )

    def run_submit_job(self, rw, header, message):
        """Submit cluster job according to message."""

        parameters = rw.recipe_step["parameters"]
        cluster_params = JobSubmissionParameters(**parameters.get("cluster", {}))

        if not isinstance(cluster_params.commands, str):
            cluster_params.commands = "\n".join(cluster_params.commands)

        if "recipefile" in parameters:
            recipefile = parameters["recipefile"]
            try:
                Path(recipefile).parent.mkdir(parents=True, exist_ok=True)
            except OSError:
                self.log.error(f"Cannot make directory for {recipefile}")
                self._transport.nack(header)
                return
            self.log.debug("Writing recipe to %s", recipefile)
            cluster_params.commands = cluster_params.commands.replace(
                "$RECIPEFILE", recipefile
            )
            with open(recipefile, "w") as fh:
                fh.write(rw.recipe.pretty())
        if "recipeenvironment" in parameters:
            recipeenvironment = parameters["recipeenvironment"]
            try:
                Path(recipeenvironment).parent.mkdir(parents=True, exist_ok=True)
            except OSError:
                self.log.error(f"Cannot make directory for {recipeenvironment}")
                self._transport.nack(header)
                return
            self.log.debug("Writing recipe environment to %s", recipeenvironment)
            cluster_params.commands = cluster_params.commands.replace(
                "$RECIPEENV", recipeenvironment
            )
            with open(recipeenvironment, "w") as fh:
                json.dump(
                    rw.environment, fh, sort_keys=True, indent=2, separators=(",", ": ")
                )
        if "recipewrapper" in parameters:
            recipewrapper = parameters["recipewrapper"]
            try:
                Path(recipewrapper).parent.mkdir(parents=True, exist_ok=True)
            except OSError:
                self.log.error(f"Cannot make directory for {recipewrapper}")
                self._transport.nack(header)
                return
            self.log.debug("Storing serialized recipe wrapper in %s", recipewrapper)
            cluster_params.commands = cluster_params.commands.replace(
                "$RECIPEWRAP", recipewrapper
            )
            with open(recipewrapper, "w") as fh:
                json.dump(
                    {
                        "recipe": rw.recipe.recipe,
                        "recipe-pointer": rw.recipe_pointer,
                        "environment": rw.environment,
                        "recipe-path": rw.recipe_path,
                        "payload": rw.payload,
                    },
                    fh,
                    indent=2,
                    separators=(",", ": "),
                )

        if "workingdir" not in parameters or not parameters["workingdir"].startswith(
            "/"
        ):
            self.log.error(
                "No absolute working directory specified. Will not run cluster job"
            )
            self._transport.nack(header)
            return
        working_directory = Path(parameters["workingdir"])
        try:
            working_directory.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            self.log.error(
                "Could not create working directory: %s", str(e), exc_info=True
            )
            self._transport.nack(header)
            return

        submit_to_scheduler = self.schedulers.get(cluster_params.scheduler)

        jobnumber = submit_to_scheduler(
            cluster_params, working_directory, self.log, zc=self.config
        )
        if not jobnumber:
            self._transport.nack(header)
            return

        # Conditionally acknowledge receipt of the message
        txn = self._transport.transaction_begin(subscription_id=header["subscription"])
        self._transport.ack(header, transaction=txn)

        # Send results onwards
        rw.set_default_channel("job_submitted")
        rw.send({"jobid": jobnumber}, transaction=txn)

        # Commit transaction
        self._transport.transaction_commit(txn)
        self.log.info(f"Submitted job {jobnumber} to {cluster_params.cluster}")



from __future__ import annotations

import datetime
import string
import time
from collections import ChainMap
from pathlib import Path
from typing import Optional

import htcondor
import workflows.recipe
from pydantic import BaseModel, Field, ValidationError, validator
from workflows.services.common_service import CommonService


class DenoiseParameters(BaseModel):
    volume: str = Field(..., min_length=1)
    output: Optional[str] = None  # volume directory
    suffix: Optional[str] = None  # ".denoised"
    model: Optional[str] = None  # "unet-3d"
    even_train_path: Optional[str] = None
    odd_train_path: Optional[str] = None
    n_train: Optional[int] = None  # 1000
    n_test: Optional[int] = None  # 200
    crop: Optional[int] = None  # 96
    base_kernel_width: Optional[int] = None  # 11
    optim: Optional[str] = None  # "adagrad"
    lr: Optional[float] = None  # 0.001
    criteria: Optional[str] = None  # "L2"
    momentum: Optional[float] = None  # 0.8
    batch_size: Optional[int] = None  # 10
    num_epochs: Optional[int] = None  # 500
    weight_decay: Optional[int] = None  # 0
    save_interval: Optional[int] = None  # 10
    save_prefix: Optional[str] = None
    num_workers: Optional[int] = None  # 1
    num_threads: Optional[int] = None  # 0
    gaussian: Optional[int] = None  # 0
    patch_size: Optional[int] = None  # 96
    patch_padding: Optional[int] = None  # 48
    device: Optional[int] = None  # -2

    @validator("model")
    def saved_models(cls, v):
        if v not in ["unet-3d-10a", "unet-3d-20a", "unet-3d"]:
            raise ValueError("Model must be one of unet-3d-10a, unet-3d-20a, unet-3d")
        return v

    @validator("optim")
    def optimizers(cls, v):
        if v not in ["adam", "adagrad", "sgd"]:
            raise ValueError("Optimizer must be one of adam, adagrad, sgd")
        return v

    @validator("criteria")
    def training_criteria(cls, v):
        if v not in ["L1", "L2"]:
            raise ValueError("Optimizer must be one of L1, L2")
        return v


class ChainMapWithReplacement(ChainMap):
    def __init__(self, *maps, substitutions=None) -> None:
        super().__init__(*maps)
        self._substitutions = substitutions

    def __getitem__(self, k):
        v = super().__getitem__(k)
        if self._substitutions and isinstance(v, str) and "$" in v:
            template = string.Template(v)
            return template.substitute(**self._substitutions)
        return v


class DenoiseIris(CommonService):
    """
    A service for denoising cryoEM tomograms using Topaz
    """

    # Human readable service name
    _service_name = "Denoise"

    # Logger name
    _logger_name = "cryoemservices.services.denoise_iris"

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Denoise service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "denoise",
            self.denoise,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def denoise(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"], "output": None}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                d_params = DenoiseParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                d_params = DenoiseParameters(**{**rw.recipe_step.get("parameters", {})})
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Denoise parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        command = ["topaz", "denoise3d", str(Path(d_params.volume).name)]

        denoise_flags = {
            "output": "-o",
            "suffix": "--suffix",
            "model": "-m",
            "even_train_path": "-a",
            "odd_train_path": "-b",
            "n_train": "--N-train",
            "n_test": "--N-test",
            "crop": "-c",
            "base_kernel_width": "--base-kernel-width",
            "optim": "--optim",
            "lr": "--lr",
            "criteria": "--criteria",
            "momentum": "--momentum",
            "batch_size": "--batch-size",
            "num_epochs": "--num-epochs",
            "weight_decay": "-w",
            "save_interval": "--save-interval",
            "num_workers": "--num-workers",
            "save_prefix": "--save-prefix",
            "num_threads": "-j",
            "gaussian": "-g",
            "patch_size": "-s",
            "patch_padding": "-p",
            "device": "-d",
        }

        for k, v in d_params.dict().items():
            if v and (k in denoise_flags):
                command.extend((denoise_flags[k], str(v)))

        suffix = str(Path(d_params.volume).suffix)
        alignment_output_dir = str(Path(d_params.volume).parent)
        denoised_file = str(Path(d_params.volume).stem) + ".denoised" + suffix
        denoised_full_path = str(Path(d_params.volume).parent) + "/" + denoised_file

        self.log.info(f"Running Topaz {command}")
        self.log.info(f"Input: {d_params.volume} Output: {denoised_full_path}")

        # Set-up condor config
        htcondor.param["RELEASE_DIR"] = "/usr"
        htcondor.param["LOCAL_DIR"] = "/var"
        htcondor.param["RUN"] = "/var/run/condor"
        htcondor.param["LOG"] = "/var/log/condor"
        htcondor.param["LOCK"] = "/var/lock/condor"
        htcondor.param["SPOOL"] = "/var/lib/condor/spool"
        htcondor.param["EXECUTE"] = "/var/lib/condor/execute"
        htcondor.param["BIN"] = "/usr/bin"
        htcondor.param["LIB"] = "/usr/lib64/condor"
        htcondor.param["INCLUDE"] = "/usr/include/condor"
        htcondor.param["SBIN"] = "/usr/sbin"
        htcondor.param["LIBEXEC"] = "/usr/libexec/condor"
        htcondor.param["SHARE"] = "/usr/share/condor"
        htcondor.param["PROCD_ADDRESS"] = "/var/run/condor/procd_pipe"
        htcondor.param[
            "JAVA_CLASSPATH_DEFAULT"
        ] = "/usr/share/condor /usr/share/condor/scimark2lib.jar ."
        htcondor.param["CONDOR_HOST"] = "pool-gpu-htcondor-manager.diamond.ac.uk"
        htcondor.param["COLLECTOR_HOST"] = "pool-gpu-htcondor-manager.diamond.ac.uk"
        htcondor.param["ALLOW_READ"] = "*"
        htcondor.param["ALLOW_WRITE"] = "*"
        htcondor.param["ALLOW_NEGOTIATOR"] = "*"
        htcondor.param["ALLOW_DAEMON"] = "*"
        htcondor.param["SEC_DEFAULT_AUTHENTICATION_METHODS"] = "FS_REMOTE, PASSWORD"
        htcondor.param[
            "SEC_WRITE_AUTHENTICATION_METHODS"
        ] = "FS_REMOTE, PASSWORD, ANONYMOUS"
        htcondor.param[
            "SEC_READ_AUTHENTICATION_METHODS"
        ] = "FS_REMOTE, PASSWORD, ANONYMOUS"
        htcondor.param["FS_REMOTE_DIR"] = "/dls/tmp/htcondor"

        output_file = str(Path(d_params.volume).with_suffix("")) + "_denoise_iris_out"
        error_file = str(Path(d_params.volume).with_suffix("")) + "_denoise_iris_error"
        log_file = str(Path(d_params.volume).with_suffix("")) + "_denoise_iris_log"
        self.log.info(f"Log file: {log_file}")

        try:
            at_job = htcondor.Submit(
                {
                    "executable": "/dls/ebic/data/staff-scratch/murfey/topaz.sh",
                    "arguments": "\"'$(input_args)'\"",  # needs to be in single quotes to be interpreted as one command
                    "output": "$(output_file)",
                    "error": "$(error_file)",
                    "log": "$(log_file)",
                    "request_gpus": "1",
                    "request_memory": "15000",
                    "request_disk": "10240",
                    "should_transfer_files": "yes",
                    "transfer_input_files": "$(volume)",
                    "transfer_output_files": "$(output_files)",
                    "initial_dir": "$(initial_dir)",
                }
            )
        except Exception:
            self.log.warn("Couldn't connect submitter for denoising")
            rw.transport.nack(header)
            return

        itemdata = [
            {
                "input_args": " ".join(command),
                "volume": d_params.volume,
                "output_file": output_file,
                "error_file": error_file,
                "log_file": log_file,
                "output_files": denoised_file,  # denoised_file,
                "initial_dir": alignment_output_dir,
            }
        ]

        coll = htcondor.Collector(htcondor.param["COLLECTOR_HOST"])
        schedd_ad = coll.locate(htcondor.DaemonTypes.Schedd)
        schedd = htcondor.Schedd(schedd_ad)
        job = schedd.submit(at_job, itemdata=iter(itemdata))
        cluster_id = job.cluster()
        self.log.info(f"Submitting to Iris, ID: {str(cluster_id)}")

        res = 1
        start_time = datetime.datetime.now()
        while res:
            try:
                res = schedd.query(
                    constraint="ClusterId=={}".format(cluster_id),
                    projection=["JobStatus"],
                )[0]["JobStatus"]
            except IndexError:
                break
            if res == 12:
                schedd.act(htcondor.JobAction.Remove, f"ClusterId == {cluster_id}")
                self.log.warning("Denoising run failed on IRIS")
                rw.transport.nack(header)
                return
            time.sleep(10)
            if (datetime.datetime.now() - start_time).seconds > 20 * 60:
                # Abort if duration over 20 minutes
                schedd.act(htcondor.JobAction.Remove, f"ClusterId == {cluster_id}")
                self.log.warning("Denoising run timed out on IRIS")
                rw.transport.nack(header)
                return

        # Forward results to images service
        self.log.info(f"Sending to images service {d_params.volume}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="images",
                message={
                    "image_command": "mrc_central_slice",
                    "file": denoised_full_path,
                },
            )
            rw.transport.send(
                destination="movie",
                message={
                    "image_command": "mrc_to_apng",
                    "file": denoised_full_path,
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "image_command": "mrc_central_slice",
                    "file": denoised_full_path,
                },
            )
            rw.send_to(
                "movie",
                {
                    "image_command": "mrc_to_apng",
                    "file": denoised_full_path,
                },
            )

        self.log.info(f"Done denoising for {d_params.volume}")
        rw.transport.ack(header)
        return



from __future__ import annotations

import json
import os
import re
import shutil
import subprocess
import time
from pathlib import Path
from typing import Literal, Optional

import workflows.recipe
import yaml
from gemmi import cif
from icebreaker import ice_groups, icebreaker_equalize_multi, icebreaker_icegroups_multi
from icebreaker.five_figures import single_mic_5fig
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from cryoemservices.util.spa_relion_service_options import RelionServiceOptions


class IceBreakerParameters(BaseModel):
    input_micrographs: str = Field(..., min_length=1)
    input_particles: Optional[str] = None
    output_path: str = Field(..., min_length=1)
    icebreaker_type: str = Literal[
        "micrographs", "enhancecontrast", "summary", "particles"
    ]
    cpus: int = 10
    total_motion: float = 0
    early_motion: float = 0
    late_motion: float = 0
    mc_uuid: int
    submit_to_slurm: bool = True
    relion_options: RelionServiceOptions


class IceBreaker(CommonService):
    """
    A service that runs the IceBreaker micrographs job
    """

    # Human readable service name
    _service_name = "IceBreaker"

    # Logger name
    _logger_name = "cryoemservices.services.icebreaker"

    # Job name
    job_type = "icebreaker.micrograph_analysis"

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("IceBreaker service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "icebreaker",
            self.icebreaker,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def icebreaker(self, rw, header: dict, message: dict):
        """
        Main function which interprets received messages, runs icebreaker
        and sends messages to the ispyb and image services
        """

        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                icebreaker_params = IceBreakerParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                icebreaker_params = IceBreakerParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"IceBreaker parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        # IceBreaker requires running in the job directory
        project_dir = Path(icebreaker_params.output_path).parent.parent
        if not Path(icebreaker_params.output_path).exists():
            Path(icebreaker_params.output_path).mkdir(parents=True)
        os.chdir(icebreaker_params.output_path)
        mic_from_project = Path(icebreaker_params.input_micrographs).relative_to(
            project_dir
        )
        micrograph_name = Path(mic_from_project.name)

        self.log.info(
            f"Type: {icebreaker_params.icebreaker_type} "
            f"Input: {icebreaker_params.input_micrographs} "
            f"Output: {icebreaker_params.output_path}"
        )
        this_job_type = f"{self.job_type}.{icebreaker_params.icebreaker_type}"
        summary_results = []

        icebreaker_success = False
        if icebreaker_params.icebreaker_type in ["micrographs", "enhancecontrast"]:
            # Create the temporary working directory
            icebreaker_tmp_dir = Path(f"IB_tmp_{micrograph_name.stem}")
            if icebreaker_tmp_dir.is_dir():
                self.log.warning(
                    f"Directory {icebreaker_tmp_dir} already exists - now removing it"
                )
                shutil.rmtree(icebreaker_tmp_dir)
            icebreaker_tmp_dir.mkdir()
            (icebreaker_tmp_dir / mic_from_project.name).symlink_to(
                icebreaker_params.input_micrographs
            )

            mic_path_parts = list(mic_from_project.parts)
            mic_dir_from_job = "/".join(mic_path_parts[2:-1])
            Path(mic_dir_from_job).mkdir(parents=True, exist_ok=True)

            # Run the icebreaker flattening or grouping functions
            if icebreaker_params.icebreaker_type == "micrographs":
                (icebreaker_tmp_dir / "grouped").mkdir()
                icebreaker_icegroups_multi.multigroup(
                    icebreaker_tmp_dir / mic_from_project.name
                )
                (
                    icebreaker_tmp_dir
                    / "grouped"
                    / f"{micrograph_name.stem}_grouped.mrc"
                ).rename(f"{mic_dir_from_job}/{micrograph_name.stem}_grouped.mrc")
            else:
                (icebreaker_tmp_dir / "flattened").mkdir()
                icebreaker_equalize_multi.multigroup(
                    icebreaker_tmp_dir / mic_from_project.name
                )
                (
                    icebreaker_tmp_dir
                    / "flattened"
                    / f"{micrograph_name.stem}_flattened.mrc"
                ).rename(f"{mic_dir_from_job}/{micrograph_name.stem}_flattened.mrc")
            shutil.rmtree(icebreaker_tmp_dir)

            # Create the command this replicates
            command = [
                "ib_job",
                "--j",
                str(icebreaker_params.cpus),
                "--single_mic",
                str(mic_from_project),
                "--o",
                icebreaker_params.output_path,
            ]
            if icebreaker_params.icebreaker_type == "micrographs":
                command.extend(["--mode", "group"])
            else:
                command.extend(["--mode", "flatten"])
        elif icebreaker_params.icebreaker_type == "summary":
            # Run the icebreaker five-figure function
            Path("IB_input").mkdir(exist_ok=True)
            (Path("IB_input") / mic_from_project.name).unlink(missing_ok=True)
            (Path("IB_input") / mic_from_project.name).symlink_to(
                icebreaker_params.input_micrographs
            )
            five_fig_csv = single_mic_5fig(
                str(Path(icebreaker_params.output_path) / "IB_input" / micrograph_name)
            )
            summary_results = five_fig_csv.split(",")
            if len(summary_results) != 6:
                summary_results = ["0", "0", "0", "0", "0", "0"]

            # Create the command this replicates
            command = [
                "ib_5fig",
                "--j",
                str(icebreaker_params.cpus),
                "--single_mic",
                str(mic_from_project),
                "--o",
                icebreaker_params.output_path,
            ]
        elif icebreaker_params.icebreaker_type == "particles":
            # Create the command this replicates
            command = [
                "ib_group",
                "--j",
                str(icebreaker_params.cpus),
                "--in_mics",
                str(mic_from_project),
                "--in_parts",
                str(Path(icebreaker_params.input_particles).relative_to(project_dir)),
                "--o",
                icebreaker_params.output_path,
            ]

            if not icebreaker_params.submit_to_slurm:
                # Run the icebreaker particle batch function
                try:
                    ice_groups.main(
                        icebreaker_params.input_particles,
                        icebreaker_params.input_micrographs,
                    )
                except FileNotFoundError as e:
                    self.log.warning(f"IceBreaker failed to find file: {e}")
                    rw.transport.nack(header)
                    return

                # Create a star file with the input data
                icegroups_doc = cif.Document()
                icegroups_block = icegroups_doc.add_new_block("input_files")
                loop = icegroups_block.init_loop(
                    "", ["_rlnParticles", "_rlnMicrographs"]
                )
                loop.add_row(
                    [
                        icebreaker_params.input_particles,
                        icebreaker_params.input_micrographs,
                    ]
                )
                icegroups_doc.write_file("ib_icegroups.star")
            else:
                # Run the icebreaker command and confirm it ran successfully
                icebreaker_success = self.slurm_submission(
                    command,
                    project_dir=project_dir,
                    job_dir=Path(icebreaker_params.output_path),
                    cpus=icebreaker_params.cpus,
                )
        else:
            self.log.warning(
                f"Unknown IceBreaker job type {icebreaker_params.icebreaker_type}"
            )
            rw.transport.nack(header)
            return

        # Register the icebreaker job with the node creator
        self.log.info(f"Sending {this_job_type} to node creator")
        node_creator_parameters = {
            "job_type": this_job_type,
            "input_file": icebreaker_params.input_micrographs,
            "output_file": icebreaker_params.output_path,
            "relion_options": dict(icebreaker_params.relion_options),
            "command": " ".join(command),
            "stdout": "",
            "stderr": "",
            "results": {
                "icebreaker_type": icebreaker_params.icebreaker_type,
                "total_motion": icebreaker_params.total_motion,
                "early_motion": icebreaker_params.early_motion,
                "late_motion": icebreaker_params.late_motion,
            },
        }
        if not icebreaker_success:
            node_creator_parameters["success"] = False
        else:
            node_creator_parameters["success"] = True
        if icebreaker_params.icebreaker_type == "summary":
            # Summary jobs need to send results to node creation
            node_creator_parameters["results"]["summary"] = summary_results[1:]
        if icebreaker_params.icebreaker_type == "particles":
            node_creator_parameters[
                "input_file"
            ] += f":{icebreaker_params.input_particles}"
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={"parameters": node_creator_parameters, "content": "dummy"},
            )
        else:
            rw.send_to("node_creator", node_creator_parameters)

        # End here if the command failed
        if not icebreaker_success:
            self.log.error("IceBreaker failed")
            rw.transport.nack(header)
            return

        # Forward results to next IceBreaker job
        if icebreaker_params.icebreaker_type == "micrographs":
            self.log.info("Sending to IceBreaker summary")
            next_icebreaker_params = {
                "icebreaker_type": "summary",
                "input_micrographs": str(
                    Path(
                        re.sub(
                            ".+/job[0-9]{3}/",
                            icebreaker_params.output_path,
                            str(mic_from_project),
                        )
                    ).parent
                    / mic_from_project.stem
                )
                + "_grouped.mrc",
                "relion_options": dict(icebreaker_params.relion_options),
                "mc_uuid": icebreaker_params.mc_uuid,
            }
            job_number = int(
                re.search("/job[0-9]{3}/", icebreaker_params.output_path)[0][4:7]
            )
            next_icebreaker_params["output_path"] = re.sub(
                f"IceBreaker/job{job_number:03}/",
                f"IceBreaker/job{job_number + 2:03}/",
                icebreaker_params.output_path,
            )
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="icebreaker",
                    message={"parameters": next_icebreaker_params, "content": "dummy"},
                )
            else:
                rw.send_to("icebreaker", next_icebreaker_params)

        # Send results to ispyb
        if icebreaker_params.icebreaker_type == "summary":
            ispyb_parameters = {
                "ispyb_command": "buffer",
                "buffer_lookup": {"motion_correction_id": icebreaker_params.mc_uuid},
                "buffer_command": {"ispyb_command": "insert_relative_ice_thickness"},
                "minimum": summary_results[1],
                "q1": summary_results[2],
                "median": summary_results[3],
                "q3": summary_results[4],
                "maximum": summary_results[5],
            }
            self.log.info(f"Sending to ispyb: {ispyb_parameters}")
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="ispyb_connector",
                    message={"parameters": ispyb_parameters, "content": "dummy"},
                )
            else:
                rw.send_to("ispyb_connector", ispyb_parameters)

        # Create symlink for the particle grouping jobs
        if (
            icebreaker_params.icebreaker_type == "particles"
            and Path(icebreaker_params.input_particles).name == "particles_split1.star"
        ):
            Path(project_dir / "IceBreaker/Icebreaker_group_batch_1").unlink(
                missing_ok=True
            )
            Path(project_dir / "IceBreaker/Icebreaker_group_batch_1").symlink_to(
                icebreaker_params.output_path
            )

        self.log.info(
            f"Done {this_job_type} for {icebreaker_params.input_micrographs}."
        )
        rw.transport.ack(header)

    def slurm_submission(
        self, command: list, project_dir: Path, job_dir: Path, cpus: int
    ):
        """Submit jobs to a slurm cluster via the RestAPI"""
        slurm_json_job_template = {
            "v0.0.38": {
                "name": "IceBreaker",
                "nodes": 1,
                "tasks": 1,
                "cpus_per_task": cpus,
                "memory_per_cpu": 12000,
                "time_limit": "1:00:00",
            },
            "v0.0.40": {
                "name": "IceBreaker",
                "minimum_nodes": 1,
                "maximum_nodes": 1,
                "tasks": 1,
                "cpus_per_task": cpus,
                "memory_per_node": {
                    "number": 12000,
                    "set": True,
                    "infinite": False,
                },
                "time_limit": {
                    "number": 3600,
                    "set": True,
                    "infinite": False,
                },
            },
        }
        slurm_script_template = (
            "#!/bin/bash\n"
            "echo \"$(date '+%Y-%m-%d %H:%M:%S.%3N'): running IceBreaker\"\n"
            "source /etc/profile.d/modules.sh\n"
            "module load EM/icebreaker/0.3.9\n"
        )

        try:
            # Get the configuration and token for the restAPI
            with open(os.environ["SLURM_RESTAPI_CONFIG"], "r") as f:
                slurm_rest = yaml.safe_load(f)
            user = slurm_rest["user"]
            user_home = slurm_rest["user_home"]
            with open(slurm_rest["user_token"], "r") as f:
                slurm_token = f.read().strip()
        except (KeyError, FileNotFoundError):
            self.log.error("Unable to load slurm restAPI config file and token")
            return False

        # Check the API version is one this service has been tested with
        api_version = slurm_rest["api_version"]
        print(api_version)
        if api_version not in ["v0.0.38", "v0.0.40"]:
            return subprocess.CompletedProcess(
                args="",
                returncode=1,
                stdout="".encode("utf8"),
                stderr=f"Unsupported API version {api_version}".encode("utf8"),
            )

        # Construct the json for submission
        mc_output_file = f"{job_dir}/slurm.out"
        mc_error_file = f"{job_dir}/slurm.err"
        submission_file = f"{job_dir}/slurm.json"
        slurm_config = {
            "standard_output": mc_output_file,
            "standard_error": mc_error_file,
            "current_working_directory": str(project_dir),
        }
        if api_version == "v0.0.38":
            slurm_config["environment"] = {"USER": user, "HOME": user_home}
        else:
            slurm_config["environment"] = [f"USER: {user}", f"HOME: {user_home}"]

        # Add slurm partition and cluster preferences if given
        if slurm_rest.get("partition"):
            slurm_config["partition"] = slurm_rest["partition"]
        if slurm_rest.get("partition_preference"):
            slurm_config["prefer"] = slurm_rest["partition_preference"]
        if slurm_rest.get("clusters"):
            slurm_config["clusters"] = slurm_rest["clusters"]
        # Combine this with the template for the given API version
        slurm_json_job = dict(slurm_json_job_template[api_version], **slurm_config)
        slurm_json_job["cpus_per_task"] = cpus

        # Make the script command and save the submission json
        job_command = slurm_script_template + " ".join(command)
        slurm_json = {"job": slurm_json_job, "script": job_command}
        with open(submission_file, "w") as f:
            json.dump(slurm_json, f)

        # Command to submit jobs to the restAPI
        slurm_submit_command = (
            f'curl -H "X-SLURM-USER-NAME:{user}" -H "X-SLURM-USER-TOKEN:{slurm_token}" '
            '-H "Content-Type: application/json" -X POST '
            f'{slurm_rest["url"]}/slurm/{slurm_rest["api_version"]}/job/submit '
            f"-d @{submission_file}"
        )
        slurm_submission_json = subprocess.run(
            slurm_submit_command, capture_output=True, shell=True
        )
        try:
            # Extract the job id from the submission response to use in the next query
            slurm_response = slurm_submission_json.stdout.decode("utf8", "replace")
            job_id = json.loads(slurm_response)["job_id"]
        except (json.JSONDecodeError, KeyError):
            self.log.error(
                f"Unable to submit job to {slurm_rest['url']}. The restAPI returned "
                f"{slurm_submission_json.stdout.decode('utf8', 'replace')}"
            )
            return False
        self.log.info(f"Submitted job {job_id} to Wilson. Waiting...")

        # Command to get the status of the submitted job from the restAPI
        slurm_status_command = (
            f'curl -H "X-SLURM-USER-NAME:{user}" -H "X-SLURM-USER-TOKEN:{slurm_token}" '
            '-H "Content-Type: application/json" -X GET '
            f'{slurm_rest["url"]}/slurm/{slurm_rest["api_version"]}/job/{job_id}'
        )
        slurm_job_state = "PENDING"

        # Wait until the job has a status indicating it has finished
        loop_counter = 0
        while slurm_job_state in (
            "PENDING",
            "CONFIGURING",
            "RUNNING",
            "COMPLETING",
        ):
            if loop_counter < 5:
                time.sleep(5)
            else:
                time.sleep(30)
            loop_counter += 1

            # Call the restAPI to find out the job state
            slurm_status_json = subprocess.run(
                slurm_status_command, capture_output=True, shell=True
            )
            try:
                slurm_response = slurm_status_json.stdout.decode("utf8", "replace")
                slurm_job_state = json.loads(slurm_response)["jobs"][0]["job_state"]
            except (json.JSONDecodeError, KeyError):
                print(slurm_status_command)
                self.log.error(
                    f"Unable to get status for job {job_id}. The restAPI returned "
                    f"{slurm_status_json.stdout.decode('utf8', 'replace')}"
                )
                return False

        # Read in the output then clean up the files
        self.log.info(f"Job {job_id} has finished!")
        if slurm_job_state == "COMPLETED":
            return True
        else:
            return False



from __future__ import annotations

import math
from pathlib import Path
from typing import Any, Dict

from pydantic import BaseModel, root_validator


def calculate_box_size(particle_size_pixels):
    # Use box 20% larger than particle and ensure size is even
    box_size_exact = 1.2 * particle_size_pixels
    box_size_int = int(math.ceil(box_size_exact))
    return box_size_int + box_size_int % 2


def calculate_downscaled_box_size(box_size_pix, angpix):
    for small_box_pix in (
        48,
        64,
        96,
        128,
        160,
        192,
        256,
        288,
        300,
        320,
        360,
        384,
        400,
        420,
        450,
        480,
        512,
        640,
        768,
        896,
        1024,
    ):
        # Don't go larger than the original box
        if small_box_pix > box_size_pix:
            return box_size_pix
        # If Nyquist freq. is better than 8.5 A, use this downscaled box, else step size
        small_box_angpix = angpix * box_size_pix / small_box_pix
        if small_box_angpix < 4.25:
            return small_box_pix
    # Fall back to a warning message
    return "Box size is too large!"


class RelionServiceOptions(BaseModel):
    """The parameters used by the Relion services"""

    """Parameters that Murfey will set"""
    # Pixel size in Angstroms in the input movies
    pixel_size: float = 0.885
    # Dose in electrons per squared Angstrom per frame
    dose_per_frame: float = 1.277
    # Gain-reference image in MRC format
    gain_ref: str = "Movies/gain.mrc"
    # Acceleration voltage (in kV)
    voltage: int = 300
    # Use binning=2 for super-resolution K2 movies
    motion_corr_binning: int = 1
    # eer format grouping
    eer_grouping: int = 20
    # Symmetry group
    symmetry: str = "C1"
    # Down-scale the particles upon extraction?
    downscale: bool = False
    # Run icebreaker?
    do_icebreaker_jobs: bool = True

    """Parameters used in internal calculations"""
    pixel_size_downscaled: float = 0
    # Diameter of particles picked by cryolo
    particle_diameter: float = 0
    # Box size of particles in the averaged micrographs (in pixels)
    boxsize: int = 256
    # Box size of the down-scaled particles (in pixels)
    small_boxsize: int = 64
    # Diameter of the mask used for 2D/3D classification (in Angstrom)
    mask_diameter: float = 190

    """Parameters we set differently from pipeliner defaults"""
    # Spherical aberration
    spher_aber: float = 2.7
    # Amplitude contrast (Q0)
    ampl_contrast: float = 0.1

    # Local motion-estimation patches for MotionCor2
    patch_sizes: dict = {"x": 5, "y": 5}

    # Additional arguments for RELION's Motion Correction wrapper
    motion_corr_other_args: str = "--do_at_most 200 --skip_logfile"
    # Threshold for cryolo autopicking
    cryolo_threshold: float = 0.15
    # Location of the cryolo specific files
    cryolo_config_file: str = "/dls_sw/apps/EM/crYOLO/phosaurus_models/config.json"
    cryolo_model_weights: str = (
        "/dls_sw/apps/EM/crYOLO/phosaurus_models/gmodel_phosnet_202005_N63_c17.h5"
    )

    # Fraction of classes to attempt to remove using the RELION 2D class ranker
    class2d_fraction_of_classes_to_remove: float = 0.9
    # Threshold to apply to class scores
    autoselect_min_score: float = 0.7
    # 2D classification particle batch size
    batch_size: int = 50000
    # Maximum batch size for the single batch of 3D classification
    class3d_max_size: int = 200000
    # Initial lowpass filter on 3D reference
    initial_lowpass: int = 40

    # Classification batches and iteration counts
    class2d_nr_classes: int = 50
    class2d_nr_iter: int = 25
    class3d_nr_classes: int = 4
    class3d_nr_iter: int = 25

    class Config:
        validate_assignment = True

    @root_validator(skip_on_failure=True)
    def if_particle_diameter_compute_box_sizes(cls, values):
        if values.get("particle_diameter"):
            values["mask_diameter"] = 1.1 * values["particle_diameter"]
            values["boxsize"] = calculate_box_size(
                values["particle_diameter"] / values["pixel_size"]
            )
            values["small_boxsize"] = calculate_downscaled_box_size(
                values["boxsize"], values["pixel_size"]
            )
        return values


def generate_service_options(
    relion_options: RelionServiceOptions, submission_type: str
) -> dict:
    job_options: Dict[str, Any] = {}
    queue_options = {
        "do_queue": "No",
        "qsubscript": "run_as_service",
    }

    job_options["relion.import.movies"] = {
        "angpix": relion_options.pixel_size,
        "kV": relion_options.voltage,
    }

    job_options["relion.motioncorr.own"] = {
        "dose_per_frame": relion_options.dose_per_frame,
        "fn_gain_ref": relion_options.gain_ref
        if Path(relion_options.gain_ref).exists()
        else "",
        "eer_grouping": relion_options.eer_grouping,
        "patch_x": relion_options.patch_sizes["x"],
        "patch_y": relion_options.patch_sizes["y"],
        "bin_factor": relion_options.motion_corr_binning,
        "other_args": f"{relion_options.motion_corr_other_args}",
        "nr_mpi": 4,
        "nr_threads": 10,
    }

    job_options["relion.motioncorr.motioncor2"] = {
        **job_options["relion.motioncorr.own"],
        "fn_motioncor2_exe": "MotionCor2",
        "gpu_ids": "0:1:2:3",
    }

    job_options["icebreaker.micrograph_analysis.micrographs"] = {"nr_threads": "10"}

    job_options["icebreaker.micrograph_analysis.enhancecontrast"] = {"nr_threads": "10"}

    job_options["icebreaker.micrograph_analysis.summary"] = {"nr_threads": "10"}

    job_options["icebreaker.micrograph_analysis.particles"] = {"nr_threads": "10"}

    job_options["relion.ctffind.ctffind4"] = {"nr_mpi": 40}

    job_options["cryolo.autopick"] = {
        "model_path": relion_options.cryolo_model_weights,
        "config_file": relion_options.cryolo_config_file,
        "box_size": "160",
        "confidence_threshold": relion_options.cryolo_threshold,
        "gpus": "0 1 2 3",
    }

    job_options["relion.extract"] = {
        "bg_diameter": -1,
        "extract_size": relion_options.boxsize,
        "do_rescale": relion_options.downscale,
        "rescale": relion_options.small_boxsize,
        "nr_mpi": 40,
    }

    job_options["relion.select.split"] = {
        "split_size": relion_options.batch_size,
    }

    job_options["relion.class2d.em"] = {
        "nr_classes": relion_options.class2d_nr_classes,
        "nr_iter_em": relion_options.class2d_nr_iter,
        "do_preread_images": True,
        "particle_diameter": relion_options.mask_diameter,
        "use_gpu": True,
        "gpu_ids": "",
        "nr_mpi": 5,
        "nr_threads": 8,
    }

    job_options["relion.class2d.vdam"] = {
        "nr_classes": relion_options.class2d_nr_classes,
        "nr_iter_grad": relion_options.class2d_nr_iter,
        "do_preread_images": True,
        "particle_diameter": relion_options.mask_diameter,
        "use_gpu": True,
        "gpu_ids": "0,1,2,3",
        "nr_mpi": 1,
        "nr_threads": 12,
    }

    job_options["relion.select.class2dauto"] = {
        "python_exe": "/dls_sw/apps/EM/relion/4.0/conda/bin/python",
        "rank_threshold": relion_options.autoselect_min_score,
        "other_args": "--select_min_nr_particles 500",
    }

    job_options["combine_star_files_job"] = {
        "do_split": True,
        "split_size": relion_options.batch_size,
    }

    job_options["relion.initialmodel"] = {
        "nr_classes": relion_options.class3d_nr_classes,
        "sym_name": relion_options.symmetry,
        "particle_diameter": relion_options.mask_diameter,
        "do_preread_images": True,
        "nr_pool": 10,
        "use_gpu": True,
        "gpu_ids": "0,1,2,3",
        "nr_threads": 12,
    }

    job_options["relion.class3d"] = {
        "nr_classes": relion_options.class3d_nr_classes,
        "nr_iter": relion_options.class3d_nr_iter,
        "sym_name": relion_options.symmetry,
        "ini_high": relion_options.initial_lowpass,
        "particle_diameter": relion_options.mask_diameter,
        "do_preread_images": True,
        "use_gpu": True,
        "gpu_ids": "",
        "nr_mpi": 5,
        "nr_threads": 8,
    }

    if submission_type not in ["relion.import.movies", "combine_star_files_job"]:
        job_options[submission_type].update(queue_options)
    return job_options[submission_type]


def update_relion_options(relion_options: RelionServiceOptions, new_options: dict):
    relion_options_dict = dict(relion_options)

    # Drop automatically generated parameters if particle_diameter is provided
    if new_options.get("particle_diameter"):
        if "boxsize" in new_options.keys():
            del new_options["boxsize"]
        if "small_boxsize" in new_options.keys():
            del new_options["small_boxsize"]
        if "mask_diameter" in new_options.keys():
            del new_options["mask_diameter"]

    for k, v in new_options.items():
        if (v is not None) and (k in relion_options_dict.keys()):
            relion_options_dict[k] = v
    return RelionServiceOptions.parse_obj(relion_options_dict)



from __future__ import annotations

import ast
import os.path
import subprocess
import time
from pathlib import Path
from typing import List, Optional, Union

import plotly.express as px
import workflows.recipe
import workflows.transport
from pydantic import BaseModel, Field, ValidationError, validator
from workflows.services.common_service import CommonService


class TomoParameters(BaseModel):
    stack_file: str = Field(..., min_length=1)
    path_pattern: str = None
    input_file_list: str = None
    position: Optional[str] = None
    aretomo_output_file: Optional[str] = None
    vol_z: int = 1200
    align: Optional[int] = None
    out_bin: int = 4
    tilt_axis: Optional[float] = None
    tilt_cor: int = 1
    flip_int: Optional[int] = None
    flip_vol: int = 1
    wbp: Optional[int] = None
    roi_file: list = []
    patch: Optional[int] = None
    kv: Optional[int] = None
    align_file: Optional[str] = None
    angle_file: Optional[str] = None
    align_z: Optional[int] = None
    pix_size: Optional[float] = None
    init_val: Optional[int] = None
    refine_flag: Optional[int] = None
    out_imod: int = 1
    out_imod_xf: Optional[int] = None
    dark_tol: Optional[Union[int, str]] = None
    manual_tilt_offset: Optional[float] = None

    @validator("input_file_list")
    def check_only_one_is_provided(cls, v, values):
        if not v and not values.get("path_pattern"):
            raise ValueError("input_file_list or path_pattern must be provided")
        if v and values.get("path_pattern"):
            raise ValueError(
                "Message must only include one of 'path_pattern' and 'input_file_list'."
                " Both are set or one has been set by the recipe."
            )
        return v

    @validator("input_file_list")
    def convert_to_list_of_lists(cls, v):
        file_list = None
        try:
            file_list = ast.literal_eval(
                v
            )  # if input_file_list is '' it will break here
        except Exception:
            return v
        if isinstance(file_list, list) and isinstance(file_list[0], list):
            return file_list
        else:
            raise ValueError("input_file_list is not a list of lists")

    @validator("input_file_list")
    def check_lists_are_not_empty(cls, v):
        for item in v:
            if not item:
                raise ValueError("Empty list found")
        return v


class TomoAlign(CommonService):
    """
    A service for grouping and aligning tomography tilt-series with Newstack and AreTomo
    """

    # Human readable service name
    _service_name = "TomoAlign"
    # Logger name
    _logger_name = "cryoemservices.services.tomo_align"

    # Values to extract for ISPyB
    refined_tilts: List[float]
    rot_centre_z_list: List[str]
    tilt_offset: float | None = None
    rot_centre_z: str | None = None
    rot: float | None = None
    mag: float | None = None
    plot_path: str | None = None
    plot_file: str | None = None
    dark_images_file: str | None = None
    imod_directory: str | None = None
    xy_proj_file: str | None = None
    xz_proj_file: str | None = None
    central_slice_file: str | None = None
    tomogram_movie_file: str | None = None
    newstack_path: str | None = None
    alignment_output_dir: str | None = None
    stack_name: str | None = None
    alignment_quality: float | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.refined_tilts = []
        self.rot_centre_z_list = []

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("TomoAlign service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "tomo_align",
            self.tomo_align,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_tomo_output(self, tomo_stdout: str):
        for line in tomo_stdout.split("\n"):
            if line.startswith("Rot center Z"):
                self.rot_centre_z_list.append(line.split()[5])
            if line.startswith("Tilt offset"):
                self.tilt_offset = float(line.split()[2].strip(","))
            if line.startswith("Best tilt axis"):
                self.alignment_quality = float(line.split()[5])

    def extract_from_aln(self, tomo_parameters):
        tomo_aln_file = None
        x_shift = []
        y_shift = []
        self.refined_tilts = []
        aln_files = list(Path(self.alignment_output_dir).glob("*.aln"))

        file_name = Path(tomo_parameters.stack_file).stem
        for aln_file in aln_files:
            if file_name in str(aln_file):
                tomo_aln_file = aln_file

        with open(tomo_aln_file) as f:
            lines = f.readlines()
            for line in lines:
                if not line.startswith("#"):
                    line_split = line.split()
                    if self.rot is None:
                        self.rot = float(line_split[1])
                    if self.mag is None:
                        self.mag = float(line_split[2])
                    x_shift.append(float(line_split[3]))
                    y_shift.append(float(line_split[4]))
                    self.refined_tilts.append(float(line_split[9]))
        fig = px.scatter(x=x_shift, y=y_shift)
        fig.write_json(self.plot_path)
        return tomo_aln_file  # not needed anywhere atm

    def tomo_align(self, rw, header: dict, message: dict):
        class MockRW:
            transport: workflows.transport.common_transport.CommonTransport

            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            print(
                "Incoming message is not a recipe message. Simple messages can be valid"
            )
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.

            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                tomo_params = TomoParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                tomo_params = TomoParameters(**{**rw.recipe_step.get("parameters", {})})
            tomo_params.pix_size = tomo_params.pix_size * 1e10
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"TomoAlign parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        def _tilt(file_list):
            return float(file_list[1])

        if tomo_params.path_pattern:
            directory = Path(tomo_params.path_pattern).parent

            input_file_list = []
            for item in directory.glob(Path(tomo_params.path_pattern).name):
                parts = str(Path(item).with_suffix("").name).split("_")
                for part in parts:
                    if "." in part:
                        input_file_list.append([str(item), part])
            tomo_params.input_file_list = input_file_list

        self.log.info(f"Input list {tomo_params.input_file_list}")
        tomo_params.input_file_list.sort(key=_tilt)

        tilt_dict: dict = {}
        for tilt in tomo_params.input_file_list:
            if not Path(tilt[0]).is_file():
                self.log.warning(f"File not found {tilt[0]}")
                rw.transport.nack(header)
            if tilt[1] not in tilt_dict:
                tilt_dict[tilt[1]] = []
            tilt_dict[tilt[1]].append(tilt[0])

        values_to_remove = []
        for item in tilt_dict:
            values = tilt_dict[item]
            if len(values) > 1:
                # sort by age and remove oldest ones
                values.sort(key=os.path.getctime)
                values_to_remove = values[1:]

        for tilt in tomo_params.input_file_list:
            if tilt[0] in values_to_remove:
                index = tomo_params.input_file_list.index(tilt)
                self.log.warning(f"Removing: {values_to_remove}")
                tomo_params.input_file_list.remove(tomo_params.input_file_list[index])

        self.alignment_output_dir = str(Path(tomo_params.stack_file).parent)
        self.stack_name = str(Path(tomo_params.stack_file).stem)

        tomo_params.aretomo_output_file = self.stack_name + "_aretomo.mrc"
        self.aretomo_output_path = (
            self.alignment_output_dir + "/" + tomo_params.aretomo_output_file
        )
        self.plot_file = self.stack_name + "_xy_shift_plot.json"
        self.plot_path = self.alignment_output_dir + "/" + self.plot_file
        self.dark_images_file = self.stack_name + "_DarkImgs.txt"
        self.xy_proj_file = self.stack_name + "_aretomo_projXY.jpeg"
        self.xz_proj_file = self.stack_name + "_aretomo_projXZ.jpeg"
        self.central_slice_file = self.stack_name + "_aretomo_thumbnail.jpeg"
        self.tomogram_movie_file = self.stack_name + "_aretomo_movie.png"
        self.newstack_path = (
            self.alignment_output_dir + "/" + self.stack_name + "_newstack.txt"
        )

        newstack_result = self.newstack(tomo_params)
        if newstack_result.returncode:
            self.log.error(
                f"Newstack failed with exitcode {newstack_result.returncode}:\n"
                + newstack_result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        tomo_params.position = str(Path(tomo_params.input_file_list[0][0]).name).split(
            "_"
        )[1]

        p = Path(self.plot_path)
        if p.is_file():
            p.chmod(0o740)

        d = Path(self.dark_images_file)
        if d.is_file():
            d.chmod(0o740)

        if tomo_params.out_imod:
            self.imod_directory = (
                self.alignment_output_dir + "/" + self.stack_name + "_aretomo_Imod"
            )

        aretomo_result = self.aretomo(tomo_params)
        if aretomo_result.returncode:
            self.log.error(
                f"AreTomo failed with exitcode {aretomo_result.returncode}:\n"
                + aretomo_result.stderr.decode("utf8", "replace")
            )
            # Update failure processing status
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="failure",
                    message="",
                )
            else:
                rw.send_to(
                    "failure",
                    "",
                )
            rw.transport.nack(header)
            return

        if tomo_params.out_imod:
            start_time = time.time()
            while not Path(self.imod_directory).is_dir():
                time.sleep(30)
                elapsed = time.time() - start_time
                if elapsed > 600:
                    self.log.warning("Timeout waiting for Imod directory")
                    break
            else:
                _f = Path(self.imod_directory)
                _f.chmod(0o750)
                for file in _f.iterdir():
                    file.chmod(0o740)

        # Extract results for ispyb

        # XY shift plot
        # Autoproc program attachment - plot
        self.extract_from_aln(tomo_params)
        if tomo_params.tilt_cor:
            try:
                self.rot_centre_z = self.rot_centre_z_list[-1]
            except IndexError:
                self.log.warning(f"No rot Z {self.rot_centre_z_list}")

        if tomo_params.pix_size:
            pix_spacing: str | None = str(tomo_params.pix_size * tomo_params.out_bin)
        else:
            pix_spacing = None
        # Forward results to ispyb

        # Tomogram (one per-tilt-series)
        ispyb_command_list = [
            {
                "ispyb_command": "insert_tomogram",
                "volume_file": tomo_params.aretomo_output_file,
                "stack_file": tomo_params.stack_file,
                "size_x": None,  # volume image size, pix
                "size_y": None,
                "size_z": None,
                "pixel_spacing": pix_spacing,
                "tilt_angle_offset": str(self.tilt_offset),
                "z_shift": self.rot_centre_z,
                "file_directory": self.alignment_output_dir,
                "central_slice_image": self.central_slice_file,
                "tomogram_movie": self.tomogram_movie_file,
                "xy_shift_plot": self.plot_file,
                "proj_xy": self.xy_proj_file,
                "proj_xz": self.xz_proj_file,
                "alignment_quality": str(self.alignment_quality),
                "store_result": "ispyb_tomogram_id",
            }
        ]

        missing_indices = []
        if Path(self.dark_images_file).is_file():
            with open(self.dark_images_file) as f:
                missing_indices = [int(i) for i in f.readlines()[2:]]
        elif self.imod_directory and Path(self.imod_directory).is_dir():
            self.dark_images_file = str(Path(self.imod_directory) / "tilt.com")
            with open(self.dark_images_file) as f:
                lines = f.readlines()
                for line in lines:
                    if line.startswith("EXCLUDELIST"):
                        numbers = line.split(" ")
                        missing_indices = [
                            int(item.replace(",", "").strip()) for item in numbers[1:]
                        ]

        im_diff = 0
        # TiltImageAlignment (one per movie)
        for im, movie in enumerate(tomo_params.input_file_list):
            if im in missing_indices:
                im_diff += 1
            else:
                try:
                    ispyb_command_list.append(
                        {
                            "ispyb_command": "insert_tilt_image_alignment",
                            "psd_file": None,  # should be in ctf table but useful, so we will insert
                            "refined_magnification": str(self.mag),
                            "refined_tilt_angle": str(self.refined_tilts[im - im_diff])
                            if self.refined_tilts
                            else None,
                            "refined_tilt_axis": str(self.rot),
                            "path": movie[0],
                        }
                    )
                except IndexError as e:
                    self.log.error(
                        f"{e} - Dark images haven't been accounted for properly"
                    )

        ispyb_parameters = {
            "ispyb_command": "multipart_message",
            "ispyb_command_list": ispyb_command_list,
        }
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb_connector", ispyb_parameters)

        # Forward results to images service
        self.log.info(f"Sending to images service {self.aretomo_output_path}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="images",
                message={
                    "image_command": "mrc_central_slice",
                    "file": self.aretomo_output_path,
                },
            )
            rw.transport.send(
                destination="movie",
                message={
                    "image_command": "mrc_to_apng",
                    "file": self.aretomo_output_path,
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "image_command": "mrc_central_slice",
                    "file": self.aretomo_output_path,
                },
            )
            rw.send_to(
                "movie",
                {
                    "image_command": "mrc_to_apng",
                    "file": self.aretomo_output_path,
                },
            )
        xy_input = (
            self.alignment_output_dir
            + "/"
            + str(Path(self.xy_proj_file).with_suffix(".mrc"))
        )
        xz_input = (
            self.alignment_output_dir
            + "/"
            + str(Path(self.xz_proj_file).with_suffix(".mrc"))
        )
        self.log.info(f"Sending to images service {xy_input}, {xz_input}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="projxy",
                message={
                    "image_command": "mrc_to_jpeg",
                    "file": xy_input,
                },
            )
            rw.transport.send(
                destination="projxz",
                message={
                    "image_command": "mrc_to_jpeg",
                    "file": xz_input,
                },
            )
        else:
            rw.send_to(
                "projxy",
                {
                    "image_command": "mrc_to_jpeg",
                    "file": xy_input,
                },
            )
            rw.send_to(
                "projxz",
                {
                    "image_command": "mrc_to_jpeg",
                    "file": xz_input,
                },
            )

        # Forward results to denoise service
        self.log.info(f"Sending to denoise service {self.aretomo_output_path}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="denoise",
                message={
                    "volume": self.aretomo_output_path,
                },
            )
        else:
            rw.send_to(
                "denoise",
                {
                    "volume": self.aretomo_output_path,
                },
            )

        # Update success processing status
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="success",
                message="",
            )
        else:
            rw.send_to(
                "success",
                "",
            )
        self.log.info(f"Done tomogram alignment for {tomo_params.stack_file}")
        rw.transport.ack(header)

    def newstack(self, tomo_parameters):
        """
        Construct file containing a list of files
        Run newstack
        """

        # Write a file with a list of .mrcs for input to Newstack
        with open(self.newstack_path, "w") as f:
            f.write(f"{len(tomo_parameters.input_file_list)}\n")
            f.write("\n0\n".join(i[0] for i in tomo_parameters.input_file_list))
            f.write("\n0\n")

        newstack_cmd = [
            "newstack",
            "-fileinlist",
            self.newstack_path,
            "-output",
            tomo_parameters.stack_file,
            "-quiet",
        ]
        self.log.info("Running Newstack")
        result = subprocess.run(newstack_cmd)
        return result

    def aretomo(self, tomo_parameters):
        """
        Run AreTomo on output of Newstack
        """
        command = ["AreTomo", "-OutMrc", self.aretomo_output_path]

        if tomo_parameters.angle_file:
            command.extend(("-AngFile", tomo_parameters.angle_file))
        else:
            command.extend(
                (
                    "-TiltRange",
                    str(tomo_parameters.input_file_list[0][1]),  # lowest tilt
                    str(tomo_parameters.input_file_list[-1][1]),
                )
            )  # highest tilt

        if tomo_parameters.manual_tilt_offset:
            command.extend(
                (
                    "-TiltCor",
                    str(tomo_parameters.tilt_cor),
                    str(tomo_parameters.manual_tilt_offset),
                )
            )
        elif tomo_parameters.tilt_cor:
            command.extend(("-TiltCor", str(tomo_parameters.tilt_cor)))

        aretomo_flags = {
            "stack_file": "-InMrc",
            "vol_z": "-VolZ",
            "out_bin": "-OutBin",
            "tilt_axis": "-TiltAxis",
            "flip_int": "-FlipInt",
            "flip_vol": "-FlipVol",
            "wbp": "-Wbp",
            "align": "-Align",
            "roi_file": "-RoiFile",
            "patch": "-Patch",
            "kv": "-Kv",
            "align_file": "-AlnFile",
            "align_z": "-AlignZ",
            "pix_size": "-PixSize",
            "init_val": "initVal",
            "refine_flag": "refineFlag",
            "out_imod": "-OutImod",
            "out_imod_xf": "-OutXf",
            "dark_tol": "-DarkTol",
        }

        for k, v in tomo_parameters.dict().items():
            if v and (k in aretomo_flags):
                command.extend((aretomo_flags[k], str(v)))

        self.log.info(f"Running AreTomo {command}")
        self.log.info(
            f"Input stack: {tomo_parameters.stack_file} \n"
            f"Output file: {self.aretomo_output_path}"
        )
        result = subprocess.run(command, capture_output=True)
        if tomo_parameters.tilt_cor:
            self.parse_tomo_output(result.stdout.decode("utf8", "replace"))
        return result



from __future__ import annotations

# import datetime
import logging
from typing import NamedTuple, Optional

import ispyb.sqlalchemy
import sqlalchemy.exc

# from sqlalchemy import delete

logger = logging.getLogger("cryoemservices.services.ispyb_buffer")
logger.setLevel(logging.INFO)


class BufferResult(NamedTuple):
    success: bool
    value: Optional[int]


def evict(*, session):
    """Throw away buffered information after a certain time.

    This needs to be run periodically to ensure the buffer tables don't
    become unnecessarily large. Information can be deleted when it is very
    unlikely that it needs to be referred to again. As buffer entries are tied
    to a single AutoProcProgram run it should be a safe bet that the buffer
    information is only of limited use after that program has completed. We
    give 30 days to deal with transient database and service problems and any
    messages stuck in the DLQ.
    """

    # Not quite clear yet how we should do this. SQLAlchemy vs stored procedure.
    # Suggested SQL something along the lines of

    # DELETE zb
    # FROM zc_ZocaloBuffer zb
    # JOIN AutoProcProgram app
    # WHERE app.autoProcProgramId = zb.AutoProcProgramId
    # AND (DATE(app.processingEndTime) < DATE_SUB(CURDATE(), INTERVAL 30 DAY)
    #      OR DATE(app.recordTimeStamp) < DATE_SUB(CURDATE(), INTERVAL 60 DAY))

    # (
    #     delete(ispyb.sqlalchemy.ZcZocaloBuffer)
    #     .where(
    #         ispyb.sqlalchemy.AutoProcProgram.autoProcProgramId
    #         == ispyb.sqlalchemy.ZcZocaloBuffer.AutoProcProgramID
    #     )
    #     .where(
    #         (
    #             ispyb.sqlalchemy.AutoProcProgram.processingEndTime
    #             < datetime.datetime.now() - datetime.timedelta(days=30)
    #         )
    #         | (
    #             ispyb.sqlalchemy.AutoProcProgram.recordTimeStamp
    #             < datetime.datetime.now() - datetime.timedelta(days=60)
    #         )
    #     )
    # )
    # session.commit()


def load(*, session, program: int, uuid: int) -> BufferResult:
    """Load an entry from the zc_ZocaloBuffer table.

    Given an AutoProcProgramID and a client-defined unique reference (uuid)
    retrieve a reference value from the database if possible.
    """
    query = (
        session.query(ispyb.sqlalchemy.ZcZocaloBuffer)
        .filter(ispyb.sqlalchemy.ZcZocaloBuffer.AutoProcProgramID == program)
        .filter(ispyb.sqlalchemy.ZcZocaloBuffer.UUID == uuid)
    )
    try:
        result = query.one()
        logger.info(
            f"buffer lookup for {program}.{uuid} succeeded (={result.Reference})"
        )
        return BufferResult(success=True, value=result.Reference)
    except sqlalchemy.exc.NoResultFound:
        logger.info(f"buffer lookup for {program}.{uuid} failed")
        return BufferResult(success=False, value=None)


def store(*, session, program: int, uuid: int, reference: int):
    """Write an entry into the zc_ZocaloBuffer table.

    The buffer table allows decoupling of the message-sending client
    and the database server-side assigned primary keys. The client defines
    a unique reference (uuid) that it will use to refer to a real primary
    key value (reference). All uuids are relative to an AutoProcProgramID
    and will be stored for a limited time based on the underlying
    AutoProcProgram record.
    """
    entry = ispyb.sqlalchemy.ZcZocaloBuffer(
        AutoProcProgramID=program,
        UUID=uuid,
        Reference=reference,
    )
    session.merge(entry)
    logger.info(f"buffering value {reference} for {program}.{uuid}")
    session.commit()



from __future__ import annotations

import os
import re
import string
import subprocess
from collections import ChainMap
from math import hypot
from pathlib import Path
from typing import Optional

import plotly.express as px
import workflows.recipe
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError, validator
from workflows.services.common_service import CommonService

from cryoemservices.util.spa_relion_service_options import (
    RelionServiceOptions,
    update_relion_options,
)


class MotionCorrParameters(BaseModel):
    movie: str = Field(..., min_length=1)
    mrc_out: str = Field(..., min_length=1)
    experiment_type: str
    pixel_size: float
    dose_per_frame: float
    use_motioncor2: bool = True
    patch_sizes: dict = {"x": 5, "y": 5}
    gpu: int = 0
    threads: int = 1
    gain_ref: str = None
    rot_gain: int = None
    flip_gain: int = None
    dark: str = None
    use_gpus: int = None
    sum_range: Optional[dict] = None
    iter: int = None
    tol: float = None
    throw: int = None
    trunc: int = None
    fm_ref: int = 0
    voltage: int = None
    fm_int_file: str = None
    init_dose: float = None
    mag: Optional[dict] = None
    motion_corr_binning: float = None
    serial: int = None
    in_suffix: str = None
    eer_sampling: int = None
    out_stack: int = None
    bft: Optional[dict] = None
    group: int = None
    defect_file: str = None
    arc_dir: str = None
    in_fm_motion: int = None
    split_sum: int = None
    dose_motionstats_cutoff: float = 4.0
    do_icebreaker_jobs: bool = True
    movie_id: int
    mc_uuid: int
    picker_uuid: int
    relion_options: Optional[RelionServiceOptions] = None
    ctf: dict = {}

    @validator("experiment_type")
    def is_spa_or_tomo(cls, experiment):
        if experiment not in ["spa", "tomography"]:
            raise ValueError("Specify an experiment type of spa or tomography.")
        return experiment

    class Config:
        ignore_extra = True


class ChainMapWithReplacement(ChainMap):
    def __init__(self, *maps, substitutions=None) -> None:
        super().__init__(*maps)
        self._substitutions = substitutions

    def __getitem__(self, k):
        v = super().__getitem__(k)
        if self._substitutions and isinstance(v, str) and "$" in v:
            template = string.Template(v)
            return template.substitute(**self._substitutions)
        return v


class MotionCorr(CommonService):
    """
    A service for motion correcting cryoEM movies using MotionCor2
    """

    # Human readable service name
    _service_name = "MotionCorr"

    # Logger name
    _logger_name = "cryoemservices.services.motioncorr"

    # Job name
    job_type = "relion.motioncorr"

    # Values to extract for ISPyB
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_shift_list = []
        self.y_shift_list = []

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Motion correction service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "motioncorr",
            self.motion_correction,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_mc2_stdout(self, mc_stdout: str):
        """
        Read the output logs of MotionCor2 to determine
        the movement of each frame
        """
        frames_line = False
        for line in mc_stdout.split("\n"):
            # Frame reading in MotionCorr 1.4.0
            if line.startswith("...... Frame"):
                line_split = line.split()
                self.x_shift_list.append(float(line_split[-2]))
                self.y_shift_list.append(float(line_split[-1]))

            # Alternative frame reading for MotionCorr 1.6.3
            if not line:
                frames_line = False
            if frames_line:
                line_split = line.split()
                self.x_shift_list.append(float(line_split[1]))
                self.y_shift_list.append(float(line_split[2]))
            if "x Shift" in line:
                frames_line = True

    def parse_relion_mc_output(self, stdout_file):
        """
        Read the output star file made by Relion to determine
        the movement of each frame
        """
        shift_cif = cif.read_file(str(stdout_file))
        shift_block = shift_cif.find_block("global_shift")
        x_shifts_str = list(shift_block.find_loop("_rlnMicrographShiftX"))
        y_shifts_str = list(shift_block.find_loop("_rlnMicrographShiftY"))
        for frame in range(len(x_shifts_str)):
            self.x_shift_list.append(float(x_shifts_str[frame]))
            self.y_shift_list.append(float(y_shifts_str[frame]))

    def motioncor2(self, command, mrc_out):
        """Run the MotionCor2 command"""
        result = subprocess.run(command, capture_output=True)
        self.parse_mc2_stdout(result.stdout.decode("utf8", "replace"))
        return result

    def relion_motioncorr(self, command, mrc_out):
        """Run Relion's owm motion correction"""
        result = subprocess.run(command, capture_output=True)
        if Path(mrc_out).with_suffix(".star").exists():
            self.parse_relion_mc_output(Path(mrc_out).with_suffix(".star"))
        else:
            self.log.error(
                f"Relion output log {Path(mrc_out).with_suffix('.star')} not found"
            )
        return result

    def motion_correction(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"], "output": None}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        parameter_map = ChainMapWithReplacement(
            message if isinstance(message, dict) else {},
            rw.recipe_step["parameters"],
            substitutions=rw.environment,
        )

        try:
            if isinstance(message, dict):
                mc_params = MotionCorrParameters(**{**dict(parameter_map), **message})
            else:
                mc_params = MotionCorrParameters(**{**dict(parameter_map)})
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Motion correction parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        # Catch any cases where the movie does not exist
        if not Path(mc_params.movie).is_file():
            self.log.warning(f"Movie {mc_params.movie} does not exist")
            rw.transport.nack(header)
            return

        # Generate the plotting paths
        drift_plot_name = str(Path(mc_params.movie).stem) + "_drift_plot.json"
        plot_path = Path(mc_params.mrc_out).parent / drift_plot_name

        # Check if this file has been run before
        if Path(mc_params.mrc_out).is_file():
            job_is_rerun = True
        else:
            job_is_rerun = False

        # Get the eer grouping out of the fractionation file
        eer_grouping = 0
        if (
            mc_params.movie.endswith(".eer")
            and mc_params.fm_int_file
            and Path(mc_params.fm_int_file).is_file()
        ):
            with open(mc_params.fm_int_file, "r") as eer_file:
                eer_values = eer_file.readline()
                try:
                    eer_grouping = eer_values.split()[1]
                except ValueError:
                    self.log.warning("Cannot read eer grouping")

        if mc_params.experiment_type == "spa":
            # Update the relion options
            mc_params.relion_options = update_relion_options(
                mc_params.relion_options, dict(mc_params)
            )
            mc_params.relion_options.eer_grouping = eer_grouping

        # Determine the input and output files
        self.log.info(f"Input: {mc_params.movie} Output: {mc_params.mrc_out}")
        if not Path(mc_params.mrc_out).parent.exists():
            Path(mc_params.mrc_out).parent.mkdir(parents=True)
        if mc_params.movie.endswith(".mrc"):
            input_flag = "-InMrc"
        elif mc_params.movie.endswith((".tif", ".tiff")):
            input_flag = "-InTiff"
        elif mc_params.movie.endswith(".eer"):
            input_flag = "-InEer"
        else:
            self.log.error(f"No input flag found for movie {mc_params.movie}")
            rw.transport.nack(header)
            return

        # Run motion correction
        if mc_params.use_motioncor2:
            # Construct the command for MotionCor2
            self.job_type = "relion.motioncorr.motioncor2"
            self.log.info("Using MotionCor2")
            command = ["MotionCor2", input_flag, mc_params.movie]
            mc2_flags = {
                "mrc_out": "-OutMrc",
                "patch_sizes": "-Patch",
                "pixel_size": "-PixSize",
                "gain_ref": "-Gain",
                "rot_gain": "-RotGain",
                "flip_gain": "-FlipGain",
                "dark": "-Dark",
                "gpu": "-Gpu",
                "use_gpus": "-UseGpus",
                "sum_range": "-SumRange",
                "iter": "-Iter",
                "tol": "-Tol",
                "throw": "-Throw",
                "trunc": "-Trunc",
                "fm_ref": "-FmRef",
                "voltage": "-Kv",
                "dose_per_frame": "-FmDose",
                "init_dose": "-InitDose",
                "mag": "-Mag",
                "motion_corr_binning": "-FtBin",
                "serial": "-Serial",
                "in_suffix": "-InSuffix",
                "eer_sampling": "-EerSampling",
                "out_stack": "-OutStack",
                "bft": "-Bft",
                "group": "-Group",
                "defect_file": "-DefectFile",
                "arc_dir": "-ArcDir",
                "in_fm_motion": "-InFmMotion",
                "split_sum": "-SplitSum",
            }
            if mc_params.movie.endswith(".eer"):
                mc2_flags["fm_int_file"] = "-FmIntFile"

            # Add values from input parameters with flags
            for k, v in mc_params.dict().items():
                if (v is not None) and (k in mc2_flags):
                    if type(v) is dict:
                        command.extend(
                            (mc2_flags[k], " ".join(str(_) for _ in v.values()))
                        )
                    else:
                        command.extend((mc2_flags[k], str(v)))
            # Run MotionCor2
            result = self.motioncor2(command, mc_params.mrc_out)

            dose_weighted = Path(mc_params.mrc_out).parent / (
                Path(mc_params.mrc_out).stem + "_DW.mrc"
            )
            if dose_weighted.is_file():
                mc_params.mrc_out = str(dose_weighted)
        else:
            # Construct the command for Relion motion correction
            self.job_type = "relion.motioncorr.own"
            self.log.info("Using Relion's own motion correction")
            os.environ["FI_PROVIDER"] = "tcp"
            command = ["relion_motion_correction", "--use_own"]
            relion_mc_flags = {
                "threads": "--j",
                "movie": "--in_movie",
                "mrc_out": "--out_mic",
                "pixel_size": "--angpix",
                "voltage": "--voltage",
                "dose_per_frame": "--dose_per_frame",
                "gain_ref": "--gainref",
                "defect_file": "--defect_file",
                "rot_gain": "--gain_rot",
                "flip_gain": "--gain_flip",
                "motion_corr_binning": "--bin_factor",
                "eer_sampling": "--eer_upsampling",
                "init_dose": "--preexposure",
                "patch_sizes": {"--patch_x": "x", "--patch_y": "y"},
                "bft": {"--bfactor": "local_motion"},
            }
            # Add values from input parameters with flags
            for param_k, param_v in mc_params.dict().items():
                if (param_v is not None) and (param_k in relion_mc_flags):
                    if type(param_v) is dict:
                        for flag_k, flag_v in relion_mc_flags[param_k].items():
                            command.extend(
                                (flag_k, str(mc_params.dict()[param_k][flag_v]))
                            )
                    else:
                        command.extend((relion_mc_flags[param_k], str(param_v)))

            # Add eer grouping if file is eer
            if eer_grouping:
                command.extend(("--eer_grouping", eer_grouping))

            # Add some standard flags
            command.extend(("--dose_weighting", "--i", "dummy"))
            # Run Relion motion correction
            result = self.relion_motioncorr(command, mc_params.mrc_out)

        # Adjust the pixel size based on the binning
        if mc_params.motion_corr_binning:
            mc_params.pixel_size *= mc_params.motion_corr_binning
            if mc_params.relion_options:
                mc_params.relion_options.pixel_size *= mc_params.motion_corr_binning

        # Confirm the command ran successfully
        if result.returncode:
            self.log.error(
                f"Motion correction of {mc_params.movie} "
                f"failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            if mc_params.experiment_type == "spa" and not job_is_rerun:
                # On spa failure send the outputs to the node creator
                node_creator_parameters = {
                    "job_type": self.job_type,
                    "input_file": mc_params.mrc_out,
                    "output_file": mc_params.mrc_out,
                    "relion_options": dict(mc_params.relion_options),
                    "command": " ".join(command),
                    "stdout": result.stdout.decode("utf8", "replace"),
                    "stderr": result.stderr.decode("utf8", "replace"),
                    "success": False,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="node_creator",
                        message={
                            "parameters": node_creator_parameters,
                            "content": "dummy",
                        },
                    )
                else:
                    rw.send_to("node_creator", node_creator_parameters)
            rw.transport.nack(header)
            return

        # Extract results for ispyb
        total_motion = 0
        early_motion = 0
        late_motion = 0
        cutoff_frame = round(
            mc_params.dose_motionstats_cutoff / mc_params.dose_per_frame
        )
        for i in range(1, len(self.x_shift_list)):
            total_motion += hypot(
                self.x_shift_list[i] - self.x_shift_list[i - 1],
                self.y_shift_list[i] - self.y_shift_list[i - 1],
            )
            if i < cutoff_frame:
                early_motion += hypot(
                    self.x_shift_list[i] - self.x_shift_list[i - 1],
                    self.y_shift_list[i] - self.y_shift_list[i - 1],
                )
            else:
                late_motion += hypot(
                    self.x_shift_list[i] - self.x_shift_list[i - 1],
                    self.y_shift_list[i] - self.y_shift_list[i - 1],
                )
        average_motion_per_frame = total_motion / len(self.x_shift_list)

        # Extract results for ispyb
        fig = px.scatter(x=self.x_shift_list, y=self.y_shift_list)
        fig.write_json(plot_path)
        snapshot_path = Path(mc_params.mrc_out).with_suffix(".jpeg")

        # Forward results to ISPyB
        ispyb_parameters = {
            "ispyb_command": "buffer",
            "buffer_command": {"ispyb_command": "insert_motion_correction"},
            "buffer_store": mc_params.mc_uuid,
            "first_frame": 1,
            "last_frame": len(self.x_shift_list),
            "total_motion": total_motion,
            "average_motion_per_frame": average_motion_per_frame,
            "drift_plot_full_path": str(plot_path),
            "micrograph_snapshot_full_path": str(snapshot_path),
            "micrograph_full_path": str(mc_params.mrc_out),
            "patches_used_x": mc_params.patch_sizes["x"],
            "patches_used_y": mc_params.patch_sizes["y"],
            "dose_per_frame": mc_params.dose_per_frame,
        }
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb_connector", ispyb_parameters)

        # If this is SPA, determine and set up the next jobs
        if mc_params.experiment_type == "spa":
            # Set up icebreaker if requested, then ctffind
            icebreaker_output = Path(
                re.sub(
                    "MotionCorr/job002/",
                    "IceBreaker/job003/",
                    mc_params.mrc_out,
                )
            )
            if mc_params.do_icebreaker_jobs and not icebreaker_output.is_file():
                # Three IceBreaker jobs: CtfFind job is MC+4
                ctf_job_number = 6

                # Both IceBreaker micrographs and flattening inherit from motioncorr
                self.log.info(
                    f"Sending to IceBreaker micrograph analysis: {mc_params.mrc_out}"
                )
                icebreaker_job003_params = {
                    "icebreaker_type": "micrographs",
                    "input_micrographs": mc_params.mrc_out,
                    "output_path": re.sub(
                        "MotionCorr/job002/.+",
                        "IceBreaker/job003/",
                        mc_params.mrc_out,
                    ),
                    "mc_uuid": mc_params.mc_uuid,
                    "relion_options": dict(mc_params.relion_options),
                    "total_motion": total_motion,
                    "early_motion": early_motion,
                    "late_motion": late_motion,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="icebreaker",
                        message={
                            "parameters": icebreaker_job003_params,
                            "content": "dummy",
                        },
                    )
                else:
                    rw.send_to("icebreaker", icebreaker_job003_params)

                self.log.info(
                    f"Sending to IceBreaker contrast enhancement: {mc_params.mrc_out}"
                )
                icebreaker_job004_params = {
                    "icebreaker_type": "enhancecontrast",
                    "input_micrographs": mc_params.mrc_out,
                    "output_path": re.sub(
                        "MotionCorr/job002/.+",
                        "IceBreaker/job004/",
                        mc_params.mrc_out,
                    ),
                    "mc_uuid": mc_params.mc_uuid,
                    "relion_options": dict(mc_params.relion_options),
                    "total_motion": total_motion,
                    "early_motion": early_motion,
                    "late_motion": late_motion,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="icebreaker",
                        message={
                            "parameters": icebreaker_job004_params,
                            "content": "dummy",
                        },
                    )
                else:
                    rw.send_to("icebreaker", icebreaker_job004_params)
            elif mc_params.do_icebreaker_jobs and icebreaker_output.is_file():
                # On a rerun, skip IceBreaker jobs but mark the CtfFind job as MC+4
                ctf_job_number = 6
            else:
                # No IceBreaker jobs: CtfFind job is MC+1
                ctf_job_number = 3
            mc_params.ctf["output_image"] = str(
                Path(
                    mc_params.mrc_out.replace(
                        "MotionCorr/job002", f"CtfFind/job{ctf_job_number:03}"
                    )
                ).with_suffix(".ctf")
            )
            mc_params.ctf["relion_options"] = dict(mc_params.relion_options)
            mc_params.ctf["amplitude_contrast"] = mc_params.relion_options.ampl_contrast
        else:
            mc_params.ctf["output_image"] = mc_params.mrc_out.replace(
                "MotionCorr", "CTF"
            ).replace("_motion_corrected", "_ctf")

        # Forward results to ctffind (in both SPA and tomography)
        self.log.info(f"Sending to ctf: {mc_params.mrc_out}")
        mc_params.ctf["experiment_type"] = mc_params.experiment_type
        mc_params.ctf["input_image"] = mc_params.mrc_out
        mc_params.ctf["mc_uuid"] = mc_params.mc_uuid
        mc_params.ctf["picker_uuid"] = mc_params.picker_uuid
        mc_params.ctf["pixel_size"] = mc_params.pixel_size
        if isinstance(rw, MockRW):
            rw.transport.send(  # type: ignore
                destination="ctffind",
                message={"parameters": mc_params.ctf, "content": "dummy"},
            )
        else:
            rw.send_to("ctffind", mc_params.ctf)

        # Forward results to images service
        self.log.info(f"Sending to images service {mc_params.mrc_out}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="images",
                message={
                    "image_command": "mrc_to_jpeg",
                    "file": mc_params.mrc_out,
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "image_command": "mrc_to_jpeg",
                    "file": mc_params.mrc_out,
                },
            )

        # If this is a new SPA run, send the results to be processed by the node creator
        if mc_params.experiment_type == "spa" and not job_is_rerun:
            # As this is the entry point we need to import the file to the project
            self.log.info("Sending relion.import.movies to node creator")
            project_dir = Path(
                re.search(".+/job[0-9]{3}/", mc_params.mrc_out)[0]
            ).parent.parent
            import_movie = (
                project_dir
                / "Import/job001"
                / Path(mc_params.mrc_out)
                .relative_to(project_dir / "MotionCorr/job002")
                .parent
                / Path(mc_params.movie).name
            )
            if not import_movie.parent.is_dir():
                import_movie.parent.mkdir(parents=True)
            import_movie.unlink(missing_ok=True)
            import_movie.symlink_to(mc_params.movie)
            import_parameters = {
                "job_type": "relion.import.movies",
                "input_file": str(mc_params.movie),
                "output_file": str(import_movie),
                "relion_options": dict(mc_params.relion_options),
                "command": "",
                "stdout": "",
                "stderr": "",
            }
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={"parameters": import_parameters, "content": "dummy"},
                )
            else:
                rw.send_to("node_creator", import_parameters)

            # Then register the motion correction job with the node creator
            self.log.info(f"Sending {self.job_type} to node creator")
            node_creator_parameters = {
                "job_type": self.job_type,
                "input_file": str(import_movie),
                "output_file": mc_params.mrc_out,
                "relion_options": dict(mc_params.relion_options),
                "command": " ".join(command),
                "stdout": result.stdout.decode("utf8", "replace"),
                "stderr": result.stderr.decode("utf8", "replace"),
                "results": {
                    "total_motion": total_motion,
                    "early_motion": early_motion,
                    "late_motion": late_motion,
                },
            }
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={"parameters": node_creator_parameters, "content": "dummy"},
                )
            else:
                rw.send_to("node_creator", node_creator_parameters)

        # Register completion with Murfey if this is tomography
        if mc_params.experiment_type == "tomography":
            self.log.info("Sending to Murfey")
            if isinstance(rw, MockRW):
                rw.transport.send(
                    "murfey_feedback",
                    {
                        "register": "motion_corrected",
                        "movie": mc_params.movie,
                        "mrc_out": mc_params.mrc_out,
                        "movie_id": mc_params.movie_id,
                    },
                )
            else:
                rw.send_to(
                    "murfey_feedback",
                    {
                        "register": "motion_corrected",
                        "movie": mc_params.movie,
                        "mrc_out": mc_params.mrc_out,
                        "movie_id": mc_params.movie_id,
                    },
                )

        self.log.info(f"Done {self.job_type} for {mc_params.movie}.")
        rw.transport.ack(header)
        self.x_shift_list = []
        self.y_shift_list = []



from __future__ import annotations

import argparse
import json
from pathlib import Path

import workflows.transport.pika_transport as pt
from workflows.recipe import RecipeWrapper


def run():
    parser = argparse.ArgumentParser(
        description="Resubmit a failed zocalo wrapper script using the .recipewrap file"
    )
    parser.add_argument(
        "-w",
        "--wrapper",
        help="Location of the .recipewrap wrapper file to resubmit",
        dest="wrapper",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Transport configuration file for connecting to the message broker",
        dest="config",
        required=True,
    )
    args = parser.parse_args()

    if not Path(args.wrapper).is_file():
        print(f"{args.wrapper} cannot be found")
        return
    if not Path(args.config).is_file():
        print(f"{args.config} cannot be found")
        return

    # Connect to the message transport
    transport = pt.PikaTransport()
    transport.load_configuration_file(args.config)
    transport.connect()

    # Load and submit the wrapper part of the recipe
    with open(args.wrapper, "r") as wrap:
        recipe = json.load(wrap)
    rw = RecipeWrapper(message=recipe, transport=transport)
    rw._send_to_destination(rw.recipe_pointer, None, rw.payload, {})


if __name__ == "__main__":
    run()



from __future__ import annotations

import time
from typing import Any, Callable, Dict, NamedTuple, Protocol

import workflows.recipe
from importlib_metadata import entry_points
from workflows.services.common_service import CommonService


class _CallableParameter(Protocol):
    def __call__(self, key: str, default: Any = ...) -> Any:
        ...


class PluginInterface(NamedTuple):
    rw: workflows.recipe.wrapper.RecipeWrapper
    parameters: _CallableParameter
    message: Dict[str, Any]


class Images(CommonService):
    """
    A service that generates images and thumbnails.
    Plugin functions can be registered under the entry point
    'cryoemservices.services.images.plugins'. The contract is that a plugin function
    takes a single argument of type PluginInterface, and returns a truthy value
    to acknowledge success, and a falsy value to reject the related message.
    If a falsy value is returned that is not False then, additionally, an error
    is logged.
    Functions may choose to return a list of files that were generated, but
    this is optional at this time.
    """

    # Human readable service name
    _service_name = "Images"

    # Logger name
    _logger_name = "cryoemservices.services.images"

    # Dictionary to contain functions from plugins
    image_functions: dict[str, Callable]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.image_functions = {}

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Image service starting")
        self.image_functions.update(
            {
                e.name: e.load()
                for e in entry_points(group="cryoemservices.services.images.plugins")
            }
        )
        workflows.recipe.wrap_subscribe(
            self._transport,
            "images",
            self.image_call,
            acknowledgement=True,
            log_extender=self.extend_log,
        )

    def image_call(self, rw, header, message):
        """Pass incoming message to the relevant plugin function."""

        def parameters(key: str, default=None):
            if isinstance(message, dict) and message.get(key):
                return message[key]
            return rw.recipe_step.get("parameters", {}).get(key, default)

        command = parameters("image_command")
        if command not in self.image_functions:
            self.log.error(f"Unknown command: {command!r}")
            rw.transport.nack(header)
            return

        start = time.perf_counter()
        try:
            result = self.image_functions[command](
                PluginInterface(rw, parameters, message)
            )
        except (PermissionError, FileNotFoundError) as e:
            self.log.error(f"Command {command!r} raised {e}", exc_info=True)
            rw.transport.nack(header)
            return
        runtime = time.perf_counter() - start

        if result:
            self.log.info(f"Command {command!r} completed in {runtime:.1f} seconds")
            rw.transport.ack(header)
        elif result is False:
            # The assumption here is that if a function returns explicit
            # 'False' then it has already taken care of logging, so we
            # don't need yet another log record.
            rw.transport.nack(header)
        else:
            self.log.error(
                f"Command {command!r} returned {result!r} after {runtime:.1f} seconds"
            )
            rw.transport.nack(header)



from __future__ import annotations

import re
from pathlib import Path
from typing import Callable, Dict

from gemmi import cif

from cryoemservices.util.spa_relion_service_options import RelionServiceOptions

NODE_PARTICLESDATA = "ParticlesData"


def get_optics_table(
    relion_options: RelionServiceOptions, particle: bool = False, im_size: int = 0
):
    """
    Create the optics table for micrograph or particle star files.
    Particle files contain additional rows describing the extracted images.
    """
    output_cif = cif.Document()
    data_optics = output_cif.add_new_block("optics")

    scaled_pixel_size = (
        str(relion_options.pixel_size_downscaled)
        if relion_options.pixel_size_downscaled
        else str(relion_options.pixel_size)
    )

    optics_columns = [
        "OpticsGroupName",
        "OpticsGroup",
        "MicrographOriginalPixelSize",
        "Voltage",
        "SphericalAberration",
        "AmplitudeContrast",
    ]
    if particle:
        optics_columns.extend(
            [
                "ImagePixelSize",
                "ImageSize",
                "ImageDimensionality",
                "CtfDataAreCtfPremultiplied",
            ]
        )
    else:
        optics_columns.append("MicrographPixelSize")

    optics_values = [
        "opticsGroup1",
        "1",
        str(relion_options.pixel_size),
        str(relion_options.voltage),
        str(relion_options.spher_aber),
        str(relion_options.ampl_contrast),
    ]
    if particle:
        optics_values.extend([scaled_pixel_size, str(im_size), "2", "0"])
    else:
        optics_values.append(scaled_pixel_size)

    optics_loop = data_optics.init_loop("_rln", optics_columns)
    optics_loop.add_row(optics_values)
    return output_cif


def _import_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Import jobs save a list of all micrographs"""
    star_file = job_dir / "movies.star"
    added_line = [str(output_file), "1"]

    # Read and append to the existing output file, or otherwise create one
    if not star_file.exists():
        output_cif = get_optics_table(relion_options)

        data_movies = output_cif.add_new_block("movies")
        movies_loop = data_movies.init_loop(
            "_rln", ["MicrographMovieName", "OpticsGroup"]
        )
        movies_loop.add_row(added_line)
        output_cif.write_file(str(star_file), style=cif.Style.Simple)
    else:
        with open(star_file, "a") as output_cif:
            output_cif.write(" ".join(added_line) + "\n")


def _motioncorr_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Motion correction saves a list of micrographs and their motion"""
    star_file = job_dir / "corrected_micrographs.star"
    added_line = [
        str(output_file),
        str(output_file.with_suffix(".star")),
        "1",
        str(results["total_motion"]),
        str(results["early_motion"]),
        str(results["late_motion"]),
    ]

    # Read and append to the existing output file, or otherwise create one
    if not star_file.exists():
        output_cif = get_optics_table(relion_options)

        data_movies = output_cif.add_new_block("micrographs")
        movies_loop = data_movies.init_loop(
            "_rln",
            [
                "MicrographName",
                "MicrographMetadata",
                "OpticsGroup",
                "AccumMotionTotal",
                "AccumMotionEarly",
                "AccumMotionLate",
            ],
        )
        movies_loop.add_row(added_line)
        output_cif.write_file(str(star_file), style=cif.Style.Simple)
    else:
        with open(star_file, "a") as output_cif:
            output_cif.write(" ".join(added_line) + "\n")

    # Logfile is expected but will not be made
    (star_file.parent / "logfile.pdf").touch()


def _ctffind_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Ctf estimation saves a list of micrographs and their ctf parameters"""
    star_file = job_dir / "micrographs_ctf.star"

    # Results needed in the star file are stored in a txt file with the output
    with open(output_file.with_suffix(".txt"), "r") as f:
        ctf_results = f.readlines()[-1].split()
    added_line = [
        str(input_file),
        "1",
        str(output_file.with_suffix(".ctf")) + ":mrc",
        ctf_results[1],
        ctf_results[2],
        str(abs(float(ctf_results[1]) - float(ctf_results[2]))),
        ctf_results[3],
        ctf_results[5],
        ctf_results[6],
    ]

    # Read and append to the existing output file, or otherwise create one
    if not star_file.exists():
        output_cif = get_optics_table(relion_options)

        data_movies = output_cif.add_new_block("micrographs")
        movies_loop = data_movies.init_loop(
            "_rln",
            [
                "MicrographName",
                "OpticsGroup",
                "CtfImage",
                "DefocusU",
                "DefocusV",
                "CtfAstigmatism",
                "DefocusAngle",
                "CtfFigureOfMerit",
                "CtfMaxResolution",
            ],
        )
        movies_loop.add_row(added_line)
        output_cif.write_file(str(star_file), style=cif.Style.Simple)
    else:
        with open(star_file, "a") as output_cif:
            output_cif.write(" ".join(added_line) + "\n")

    # Logfile is expected but will not be made
    (star_file.parent / "logfile.pdf").touch()


def _icebreaker_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):

    if results["icebreaker_type"] != "particles":
        # Micrograph icebreaker jobs need a file listing the completed micrographs
        with open(job_dir / "done_mics.txt", "a+") as f:
            f.write(input_file.name + "\n")

    if results["icebreaker_type"] == "micrographs":
        # Micrograph jobs save a list of micrographs and their motion
        star_file = job_dir / "grouped_micrographs.star"
        file_to_add = (
            str(
                Path(re.sub(".+/job[0-9]{3}", str(output_file), str(input_file))).parent
                / input_file.stem
            )
            + "_grouped.mrc"
        )
    elif results["icebreaker_type"] == "enhancecontrast":
        # Contrast enhancement jobs save a list of micrographs and their motion
        star_file = job_dir / "flattened_micrographs.star"
        file_to_add = (
            str(
                Path(re.sub(".+/job[0-9]{3}", str(output_file), str(input_file))).parent
                / input_file.stem
            )
            + "_flattened.mrc"
        )
    elif results["icebreaker_type"] == "summary":
        csv_file = job_dir / "five_figs_test.csv"
        if not csv_file.exists():
            with open(csv_file, "w") as f:
                f.write("path,min,q1,q2=median,q3,max\n")
        with open(csv_file, "a+") as f:
            f.write(f"{input_file}," + ",".join(results["summary"]) + "\n")
        return
    else:
        # Nothing to do for particles job
        return

    # Read and append to the existing output file, or otherwise create one
    added_line = [
        file_to_add,
        str(input_file.with_suffix(".star")),
        "1",
        str(results["total_motion"]),
        str(results["early_motion"]),
        str(results["late_motion"]),
    ]
    if not star_file.exists():
        output_cif = cif.Document()

        data_movies = output_cif.add_new_block("micrographs")
        movies_loop = data_movies.init_loop(
            "_rln",
            [
                "MicrographName",
                "MicrographMetadata",
                "OpticsGroup",
                "AccumMotionTotal",
                "AccumMotionEarly",
                "AccumMotionLate",
            ],
        )
        movies_loop.add_row(added_line)
        output_cif.write_file(str(star_file), style=cif.Style.Simple)
    else:
        with open(star_file, "a") as output_cif:
            output_cif.write(" ".join(added_line) + "\n")


def _cryolo_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Cryolo jobs save a list of micrographs and files with particle coordinates"""
    star_file = job_dir / "autopick.star"
    added_line = [str(input_file), str(output_file)]

    # Read and append to the existing output file, or otherwise create one
    if not star_file.exists():
        output_cif = cif.Document()

        data_movies = output_cif.add_new_block("coordinate_files")
        movies_loop = data_movies.init_loop(
            "_rln",
            [
                "MicrographName",
                "MicrographCoordinates",
            ],
        )
        movies_loop.add_row(added_line)
        output_cif.write_file(str(star_file), style=cif.Style.Simple)
    else:
        with open(star_file, "a") as output_cif:
            output_cif.write(" ".join(added_line) + "\n")


def _extract_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Extract jobs save a list of particle coordinates"""
    star_file = job_dir / "particles.star"

    # Read and append to the existing output file, or otherwise create one
    if not star_file.exists():
        output_cif = get_optics_table(
            relion_options, particle=True, im_size=results["box_size"]
        )

        particles_cif = cif.read_file(str(output_file))
        data_particles = particles_cif.find_block("particles")
        output_cif.add_copied_block(data_particles)
        output_cif.write_file(str(star_file), style=cif.Style.Simple)

    else:
        with open(output_file, "r") as added_cif:
            added_lines = added_cif.readlines()

        with open(star_file, "a") as output_cif:
            for new_row in added_lines:
                if new_row[:1].isdigit():
                    output_cif.write(new_row)


def _select_output_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Select jobs need no further files, but have extra nodes to add"""
    # Find and add all the output files to the node list
    split_count = int(str(output_file).split("split")[1].split(".")[0])
    split_files = {}
    for split_file in range(split_count, 0, -1):
        split_name = f"particles_split{split_file}.star"
        split_files[split_name] = [NODE_PARTICLESDATA, ["relion"]]
    return split_files


def _relion_no_extra_files(
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    """Jobs run through relion do not need any extra files written"""
    return


_output_files: Dict[str, Callable] = {
    "relion.import.movies": _import_output_files,
    "relion.motioncorr.own": _motioncorr_output_files,
    "relion.motioncorr.motioncor2": _motioncorr_output_files,
    "icebreaker.micrograph_analysis.micrographs": _icebreaker_output_files,
    "icebreaker.micrograph_analysis.enhancecontrast": _icebreaker_output_files,
    "icebreaker.micrograph_analysis.summary": _icebreaker_output_files,
    "relion.ctffind.ctffind4": _ctffind_output_files,
    "cryolo.autopick": _cryolo_output_files,
    "relion.extract": _extract_output_files,
    "relion.select.split": _select_output_files,
    "icebreaker.micrograph_analysis.particles": _icebreaker_output_files,
    "relion.class2d.em": _relion_no_extra_files,
    "relion.select.class2dauto": _relion_no_extra_files,
    "combine_star_files_job": _relion_no_extra_files,
    "relion.initialmodel": _relion_no_extra_files,
    "relion.class3d": _relion_no_extra_files,
}


def create_output_files(
    job_type: str,
    job_dir: Path,
    input_file: Path,
    output_file: Path,
    relion_options: RelionServiceOptions,
    results: dict,
):
    return _output_files[job_type](
        job_dir, input_file, output_file, relion_options, results
    )



from __future__ import annotations

import re
from pathlib import Path

import mrcfile
import numpy as np
import workflows.recipe
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from cryoemservices.util.spa_relion_service_options import (
    RelionServiceOptions,
    update_relion_options,
)


class ExtractParameters(BaseModel):
    micrographs_file: str = Field(..., min_length=1)
    coord_list_file: str = Field(..., min_length=1)
    output_file: str = Field(..., min_length=1)
    pixel_size: float
    ctf_image: str
    ctf_max_resolution: float
    ctf_figure_of_merit: float
    defocus_u: float
    defocus_v: float
    defocus_angle: float
    particle_diameter: float = 0
    boxsize: int = 256
    small_boxsize: int = 64
    batch_size: int = 50000
    norm: bool = True
    bg_radius: int = -1
    downscale: bool = False
    invert_contrast: bool = True
    confidence_threshold: float = 0
    voltage: int = 300
    relion_options: RelionServiceOptions


class Extract(CommonService):
    """
    A service for extracting particles from cryolo autopicking
    """

    # Human readable service name
    _service_name = "Extract"

    # Logger name
    _logger_name = "cryoemservices.services.extract"

    # Job name
    job_type = "relion.extract"

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Extract service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "extract",
            self.extract,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def extract(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            print(
                "Incoming message is not a recipe message. Simple messages can be valid"
            )
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return
            self.log.debug("Received a simple message")

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                extract_params = ExtractParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                extract_params = ExtractParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Extraction parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        self.log.info(
            f"Inputs: {extract_params.micrographs_file}, "
            f"{extract_params.ctf_image}, "
            f"{extract_params.coord_list_file} "
            f"Output: {extract_params.output_file}"
        )

        # Update the relion options and get box sizes
        extract_params.relion_options = update_relion_options(
            extract_params.relion_options, dict(extract_params)
        )

        # Make sure the output directory exists
        job_dir = Path(re.search(".+/job[0-9]{3}/", extract_params.output_file)[0])
        project_dir = job_dir.parent.parent
        if not Path(extract_params.output_file).parent.exists():
            Path(extract_params.output_file).parent.mkdir(parents=True)
        output_mrc_file = (
            Path(extract_params.output_file).parent
            / Path(extract_params.micrographs_file).with_suffix(".mrcs").name
        )

        # If no background radius set diameter as 75% of box
        if extract_params.bg_radius == -1:
            extract_params.bg_radius = round(
                0.375
                * (
                    extract_params.relion_options.small_boxsize
                    if extract_params.downscale
                    else extract_params.relion_options.boxsize
                )
            )

        # Find the locations of the particles
        cbox_name = Path(
            extract_params.coord_list_file.replace("STAR", "CBOX")
        ).with_suffix(".cbox")
        if cbox_name.is_file() and extract_params.confidence_threshold:
            # If a threshold is given and the CBOX file exists use the confidences
            try:
                cbox_file = cif.read_file(cbox_name)
                cbox_block = cbox_file.find_block("cryolo")

                particles_confidence = np.array(
                    cbox_block.find_loop("_Confidence"), dtype=float
                )
                particles_x_all = np.array(
                    cbox_block.find_loop("_CoordinateX"), dtype=float
                )
                particles_y_all = np.array(
                    cbox_block.find_loop("_CoordinateY"), dtype=float
                )
                particles_x = particles_x_all[
                    particles_confidence > extract_params.confidence_threshold
                ]
                particles_y = particles_y_all[
                    particles_confidence > extract_params.confidence_threshold
                ]
            except (AttributeError, OSError):
                # Catch the case of CBOX files with no particles
                particles_x = np.array([])
                particles_y = np.array([])
        else:
            # Otherwise read from the star file
            coords_file = cif.read(extract_params.coord_list_file)
            coords_block = coords_file.sole_block()
            particles_x = np.array(coords_block.find_loop("_rlnCoordinateX"))
            particles_y = np.array(coords_block.find_loop("_rlnCoordinateY"))

        # Construct the output star file
        extracted_parts_doc = cif.Document()
        extracted_parts_block = extracted_parts_doc.add_new_block("particles")
        extracted_parts_loop = extracted_parts_block.init_loop(
            "_rln",
            [
                "CoordinateX",
                "CoordinateY",
                "ImageName",
                "MicrographName",
                "OpticsGroup",
                "CtfMaxResolution",
                "CtfFigureOfMerit",
                "DefocusU",
                "DefocusV",
                "DefocusAngle",
                "CtfBfactor",
                "CtfScalefactor",
                "PhaseShift",
            ],
        )
        for particle in range(len(particles_x)):
            extracted_parts_loop.add_row(
                [
                    particles_x[particle],
                    particles_y[particle],
                    f"{particle:06}@{output_mrc_file.relative_to(project_dir)}",
                    str(Path(extract_params.micrographs_file).relative_to(project_dir)),
                    "1",
                    str(extract_params.ctf_max_resolution),
                    str(extract_params.ctf_figure_of_merit),
                    str(extract_params.defocus_u),
                    str(extract_params.defocus_v),
                    str(extract_params.defocus_angle),
                    "0.0",
                    "1.0",
                    "0.0",
                ]
            )
        extracted_parts_doc.write_file(
            extract_params.output_file, style=cif.Style.Simple
        )

        # Extraction
        with mrcfile.open(extract_params.micrographs_file) as input_micrograph:
            input_micrograph_image = np.array(input_micrograph.data, dtype=np.float32)
        image_size = np.shape(input_micrograph_image)
        output_mrc_stack = []

        for particle in range(len(particles_x)):
            # Pixel locations are from bottom left, need to flip the image later
            pixel_location_x = round(float(particles_x[particle]))
            pixel_location_y = round(float(particles_y[particle]))

            # Extract the particle image and pad the edges if it is not square
            x_left_pad = 0
            x_right_pad = 0
            y_top_pad = 0
            y_bot_pad = 0

            extract_width = round(extract_params.relion_options.boxsize / 2)

            x_left = pixel_location_x - extract_width
            if x_left < 0:
                x_left_pad = -x_left
                x_left = 0
            x_right = pixel_location_x + extract_width
            if x_right >= image_size[1]:
                x_right_pad = x_right - image_size[1]
                x_right = image_size[1]
            y_top = pixel_location_y - extract_width
            if y_top < 0:
                y_top_pad = -y_top
                y_top = 0
            y_bot = pixel_location_y + extract_width
            if y_bot >= image_size[0]:
                y_bot_pad = y_bot - image_size[0]
                y_bot = image_size[0]

            particle_subimage = input_micrograph_image[y_top:y_bot, x_left:x_right]
            particle_subimage = np.pad(
                particle_subimage,
                ((y_bot_pad, y_top_pad), (x_left_pad, x_right_pad)),
                mode="edge",
            )

            # Flip all the values on inversion
            if extract_params.invert_contrast:
                particle_subimage = -1 * particle_subimage

            # Downscale the image size
            if extract_params.downscale:
                extract_params.relion_options.pixel_size_downscaled = (
                    extract_params.pixel_size
                    * extract_params.relion_options.boxsize
                    / extract_params.relion_options.small_boxsize
                )
                subimage_ft = np.fft.fftshift(np.fft.fft2(particle_subimage))
                deltax = (
                    subimage_ft.shape[0] - extract_params.relion_options.small_boxsize
                )
                deltay = (
                    subimage_ft.shape[1] - extract_params.relion_options.small_boxsize
                )
                particle_subimage = np.real(
                    np.fft.ifft2(
                        np.fft.ifftshift(
                            subimage_ft[
                                deltax // 2 : subimage_ft.shape[0] - deltax // 2,
                                deltay // 2 : subimage_ft.shape[1] - deltay // 2,
                            ]
                        )
                    )
                )
                extract_width = round(extract_params.relion_options.small_boxsize / 2)

            # Distance of each pixel from the centre, compared to background radius
            grid_indexes = np.meshgrid(
                np.arange(2 * extract_width),
                np.arange(2 * extract_width),
            )
            distance_from_centre = np.sqrt(
                (grid_indexes[0] - extract_width + 0.5) ** 2
                + (grid_indexes[1] - extract_width + 0.5) ** 2
            )
            bg_region = (
                distance_from_centre
                > np.ones(np.shape(particle_subimage)) * extract_params.bg_radius
            )

            # Fit background to a plane and subtract the plane from the image
            positions = [grid_indexes[0][bg_region], grid_indexes[1][bg_region]]
            # needs to create a matrix of the correct shape for  a*x + b*y + c plane fit
            if not len(positions[0]) == len(positions[1]):
                self.log.warning(
                    f"Particle image {particle} in {extract_params.micrographs_file} is not square"
                )
                continue
            data_size = len(positions[0])
            positions_matrix = np.hstack(
                (
                    np.reshape(positions[0], (data_size, 1)),
                    np.reshape(positions[1], (data_size, 1)),
                )
            )
            # this ones for c
            positions_matrix = np.hstack((np.ones((data_size, 1)), positions_matrix))
            values = particle_subimage[bg_region]
            # normal equation
            try:
                theta = np.dot(
                    np.dot(
                        np.linalg.inv(
                            np.dot(positions_matrix.transpose(), positions_matrix)
                        ),
                        positions_matrix.transpose(),
                    ),
                    values,
                )
            except np.linalg.LinAlgError:
                self.log.warning(
                    f"Could not fit image plane for particle {particle} in {extract_params.micrographs_file}"
                )
                continue
            # now we need the full grid across the image
            positions_matrix = np.hstack(
                (
                    np.reshape(grid_indexes[0], (4 * extract_width**2, 1)),
                    np.reshape(grid_indexes[1], (4 * extract_width**2, 1)),
                )
            )
            positions_matrix = np.hstack(
                (np.ones((4 * extract_width**2, 1)), positions_matrix)
            )
            plane = np.reshape(
                np.dot(positions_matrix, theta), (2 * extract_width, 2 * extract_width)
            )

            particle_subimage -= plane

            # Background normalisation
            if extract_params.norm:
                # Standardise the values using the background
                bg_mean = np.mean(particle_subimage[bg_region])
                bg_std = np.std(particle_subimage[bg_region])
                particle_subimage = (particle_subimage - bg_mean) / bg_std

            # Add to output stack
            if len(output_mrc_stack):
                output_mrc_stack = np.append(
                    output_mrc_stack, [particle_subimage], axis=0
                )
            else:
                output_mrc_stack = np.array([particle_subimage], dtype=np.float32)

        # Produce the mrc file of the extracted particles
        if extract_params.downscale:
            box_len = extract_params.relion_options.small_boxsize
            pixel_size = extract_params.relion_options.pixel_size_downscaled
        else:
            box_len = extract_params.relion_options.boxsize
            pixel_size = extract_params.relion_options.pixel_size

        particle_count = np.shape(output_mrc_stack)[0]
        self.log.info(f"Extracted {particle_count} particles")
        if particle_count > 0:
            with mrcfile.new(str(output_mrc_file), overwrite=True) as mrc:
                mrc.set_data(output_mrc_stack.astype(np.float32))
                mrc.header.mx = box_len
                mrc.header.my = box_len
                mrc.header.mz = 1
                mrc.header.cella.x = pixel_size * box_len
                mrc.header.cella.y = pixel_size * box_len
                mrc.header.cella.z = 1

        # Register the extract job with the node creator
        self.log.info(f"Sending {self.job_type} to node creator")
        node_creator_parameters = {
            "job_type": self.job_type,
            "input_file": extract_params.coord_list_file
            + ":"
            + extract_params.ctf_image,
            "output_file": extract_params.output_file,
            "relion_options": dict(extract_params.relion_options),
            "command": "",
            "stdout": "",
            "stderr": "",
            "results": {"box_size": box_len},
        }
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={"parameters": node_creator_parameters, "content": "dummy"},
            )
        else:
            rw.send_to("node_creator", node_creator_parameters)

        # Register the files needed for selection and batching
        self.log.info("Sending to particle selection")
        select_params = {
            "input_file": extract_params.output_file,
            "batch_size": extract_params.batch_size,
            "image_size": box_len,
            "relion_options": dict(extract_params.relion_options),
        }
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="select_particles",
                message={"parameters": select_params, "content": "dummy"},
            )
        else:
            rw.send_to("select_particles", select_params)

        self.log.info(f"Done {self.job_type} for {extract_params.coord_list_file}.")
        rw.transport.ack(header)
